# -*- coding: utf-8 -*-

"""
Types for closures and inner functions.
"""

from __future__ import print_function, division, absolute_import

from numba.typesystem import NumbaType
from numba.exttypes.types.extensiontype import ExtensionType

class ClosureType(NumbaType):
    """
    Type of closures and inner functions.
    """

    typename = "closure"
    argnames = ["signature", ("closure", None)]
    flags = ["object"]

    def add_scope_arg(self, scope_type):
        self.signature = self.signature.add_arg(0, scope_type)

    def __repr__(self):
        return "<closure(%s)>" % self.signature

class ClosureScopeType(ExtensionType):
    """
    Type of the enclosing scope for closures. This is always passed in as
    first argument to the function.
    """

    typename = "closure_scope"
    is_final = True

    def __init__(self, py_class, parent_scope):
        super(ClosureScopeType, self).__init__(py_class)
        self.parent_scope = parent_scope
        self.unmangled_symtab = None

        if self.parent_scope is None:
            self.scope_prefix = ""
        else:
            self.scope_prefix = self.parent_scope.scope_prefix + "0"


"""
>>> native_convert(char, -10)
-10
>>> native_convert(char, 10)
10
>>> native_convert(char, 127)
127

This doesn't work yet, we should get an error here. We don't get one because
autojit detects the int type which is natively truncated to a char.

TODO::::::::::

    >> native_convert(char, 128)

        => need exception!

>>> object_convert(char, 128)
Traceback (most recent call last):
    ...
OverflowError: value too large to convert to signed char
>>> object_convert(char, -128)
-128
>>> object_convert(char, -129)
Traceback (most recent call last):
    ...
OverflowError: value too large to convert to signed char
>>> object_convert(char, 2.9)
2

TODO:::::::::::

    Test all numeric types for overflows!

TODO:::::::::::

    Test typedef types (npy_intp, Py_uintptr_t, etc)
"""

import unittest
from numba import *

@autojit
def native_convert(dst_type, value):
    return dst_type(value)

@autojit(locals=dict(obj=object_))
def object_convert(dst_type, obj):
    return dst_type(obj)

class TestConversion(unittest.TestCase):

    def test_native_conversion(self):
        assert native_convert(char, -10) == b'\xf6'
        assert native_convert(char, 10)  == b'\n'
        assert native_convert(char, 127) == b'\x7f'

        # TODO: the below should raise an exception
        # We don't get one because autojit detects the int type which is
        # simply truncated to a char
        # native_convert(char, 128)

    def test_object_conversion(self):
        assert object_convert(char, -128) == b'\x80'
        assert object_convert(char, 2.9)  == b'\x02'

    def test_overflow(self):
        self._convert_overflow(128 , char,   'signed char')
        self._convert_overflow(-129, char,   'signed char')
        self._convert_overflow(2**31, int32, 'signed int')

    def _convert_overflow(self, value, type, typename):
        self.assertRaises(OverflowError, object_convert, type, value)
        # self.assertEqual(captured.exception.args[0],
        #                  "value too large to convert to %s" % typename)

if __name__ == "__main__":
    unittest.main()



from numba import *
from numba import error

#@autojit
def func():
    if x:
        print("hello")
    else:
        print("world")

def compile_func1():
    try:
        jit(void())(func)
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ = """
>>> compile_func1()
exception: (see below)
--------------------- Numba Encountered Errors or Warnings ---------------------
    if x:
-------^
Error ...: No global named 'x'
--------------------------------------------------------------------------------
"""

#@autojit
def func2():
    print(10[20])

def compile_func2():
    try:
        jit(void())(func2)
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ += """>>> compile_func2()
exception: (see below)
--------------------- Numba Encountered Errors or Warnings ---------------------
    print(10[20])
----------^
Error ...: object of type int cannot be indexed
--------------------------------------------------------------------------------
"""

@autojit # this often messes up line numbers
def func_decorated():
    print(10[20])

def compile_func3():
    try:
        func_decorated()
    except error.NumbaError as e:
        print("exception: %s" % e)

__doc__ += """
>>> compile_func3()
exception: (see below)
--------------------- Numba Encountered Errors or Warnings ---------------------
    print(10[20])
----------^
Error ...: object of type int cannot be indexed
--------------------------------------------------------------------------------
"""

def warn_and_error(a, b):
    print(a)
    1[2]

__doc__ += """
>>> autojit(warn=False)(warn_and_error)(1, 2)
Traceback (most recent call last):
    ...
NumbaError: (see below)
--------------------- Numba Encountered Errors or Warnings ---------------------
    1[2]
----^
Error 68:4: object of type int cannot be indexed
<BLANKLINE>
--------------------------------------------------------------------------------

>>> autojit(warnstyle='simple')(warn_and_error)(1, 2)
Traceback (most recent call last):
    ...
NumbaError: (see below)
Error ...: object of type int cannot be indexed
Warning ...: Unused argument 'b'

>>> autojit(func_decorated.py_func, warnstyle='simple')()
Traceback (most recent call last):
    ...
NumbaError: ...: object of type int cannot be indexed
"""

if __name__ == '__main__':
    import numba
    numba.testing.testmod()



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import numba
import numpy

n = 50
steps = 10000

@numba.autojit()
def calc():
    value = numpy.zeros((n,n))

    for i in range(n):
        for j in range(n):
            value[i][j] += 1.0

    return value

def leaky():
    for i in range(steps):
        value = calc()

leaky()



# -*- coding: utf-8 -*-

"""
Generate a package with IR implementations and tools.
"""

from __future__ import print_function, division, absolute_import

import os
from textwrap import dedent
from itertools import chain

from . import generator
from . import formatting
from . import astgen
from . import visitorgen
from . import naming

#------------------------------------------------------------------------
# Tools Flags
#------------------------------------------------------------------------

cython = 1

#------------------------------------------------------------------------
# Tools Resolution
#------------------------------------------------------------------------

class Tool(object):
    def __init__(self, codegens, flags=0, depends=[]):
        self.codegens = codegens
        self.flags = flags
        self.depends = depends

    def __repr__(self):
        return "Tool(codegens=[%s])" % ", ".join(map(str, self.codegens))

def resolve_tools(tool_list, mask, tools=None, seen=None):
    if tools is None:
        tools = []
        seen = set()

    for tool in tool_list:
        if not (tool.flags & mask) and tool not in seen:
            seen.add(tool)
            resolve_tools(tool.depends, mask, tools, seen)
            tools.append(tool)

    return tools

def enumerate_tools(feature_names, mask):
    tool_set = set(chain(*[features[name] for name in feature_names]))
    tools = resolve_tools(tool_set, mask)
    return tools

def enumerate_codegens(feature_names, mask):
    tools = enumerate_tools(feature_names, mask)
    codegens = list(chain(*[tool.codegens for tool in tools]))
    return codegens

#------------------------------------------------------------------------
# Tool Definitions
#------------------------------------------------------------------------

def make_codegen_dict(codegens):
    return dict((codegen.out_filename, codegen) for codegen in codegens)

all_codegens = astgen.codegens + visitorgen.codegens
gens = make_codegen_dict(all_codegens)

pxd_ast_tool        = Tool([gens[naming.nodes + ".pxd"]], flags=cython)
py_ast_tool         = Tool([gens[naming.nodes + ".py"]])

pxd_interface_tool  = Tool([gens[naming.interface + ".pxd"]], flags=cython,
                           depends=[pxd_ast_tool])
py_interface_tool   = Tool([gens[naming.interface + ".py"]],
                           depends=[py_ast_tool])

pxd_visitor_tool    = Tool([gens[naming.visitor + ".pxd"]], flags=cython,
                           depends=[pxd_interface_tool])
py_visitor_tool     = Tool([gens[naming.visitor + ".py"]],
                           depends=[py_interface_tool, pxd_visitor_tool])

pxd_transform_tool  = Tool([gens[naming.transformer + ".pxd"]], flags=cython,
                           depends=[pxd_interface_tool])
py_transformr_tool  = Tool([gens[naming.transformer + ".py"]],
                           depends=[py_interface_tool, pxd_transform_tool])

pxd_ast_tool.depends.extend([pxd_interface_tool, py_interface_tool])

#------------------------------------------------------------------------
# Feature Definitions & Entry Points
#------------------------------------------------------------------------

features = {
    'all': [py_ast_tool, py_visitor_tool, py_transformr_tool],
    'ast': [py_ast_tool],
    'visitor': [py_visitor_tool],
    'transformer': [py_transformr_tool],
}

def build_package(schema_filename, feature_names, output_dir, mask=0):
    """
    Build a package from the given schema and feature names in output_dir.

    :param mask: indicates which features to mask, e.g. specifying
                 'mask=build.cython' disables Cython support.
    """
    codegens = enumerate_codegens(feature_names, mask)
    disk_allocator = generator.generate_from_file(
        schema_filename, codegens, output_dir)

    try:
        _make_package(disk_allocator, codegens)
    finally:
        disk_allocator.close()

#------------------------------------------------------------------------
# Package Building Utilities
#------------------------------------------------------------------------

source_name = lambda fn: os.path.splitext(os.path.basename(fn))[0]

def _make_package(disk_allocator, codegens):
    _make_init(disk_allocator, codegens)

    # Make Cython dependency optional
    # disk_allocator.open_sourcefile("cython.py")

    fns = [c.out_filename for c in codegens if c.out_filename.endswith('.pxd')]
    if fns:
        _make_setup(disk_allocator, [source_name(fn) + '.py' for fn in fns])

def _make_init(disk_allocator, codegens):
    init = disk_allocator.open_sourcefile("__init__.py")
    init.write(dedent("""
        # Horrid hack to make work around circular cimports
        import os, sys
        sys.path.append(os.path.dirname(os.path.abspath(__file__)))
    """))

    for c in codegens:
        if c.out_filename.endswith('.py'):
            modname = source_name(c.out_filename)
            init.write("from %s import *\n" % modname)

def _make_setup(disk_allocator, filenames):
    setup = disk_allocator.open_sourcefile("setup.py")

    ext_modules = ["Extension('%s', ['%s'])" % (source_name(fn), fn)
                      for fn in filenames]

    setup.write(dedent("""
        from distutils.core import setup
        from Cython.Distutils import build_ext
        from Cython.Distutils.extension import Extension

        ext_modules = [
            %s
        ]

        setup(
            # ext_modules=cythonize('*.pyx'),
            ext_modules=ext_modules,
            cmdclass={'build_ext': build_ext},
        )
    """) % formatting.py_formatter.format_stats(",\n", 4, ext_modules))



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import ctypes
import inspect
from functools import partial

# from numba import llvm_types
from numba.typesystem.itypesystem import tyname
from numba import llvm_types
from numba.typesystem import itypesystem, universe
from numba.typesystem import (numba_typesystem as ts,
                              llvm_typesystem as lts,
                              ctypes_typesystem as cts)

typenames = universe.int_typenames + universe.float_typenames + ["void"]

def convert(ts1, ts2, conversion_type, typenames):
    for typename in typenames:
        t1 = getattr(ts1, tyname(typename))
        t2 = getattr(ts2, tyname(typename))
        converted = ts1.convert(conversion_type, t1)
        assert converted == t2, (str(t1), str(converted), str(t2))

#-------------------------------------------------------------------
# Numba -> LLVM
#-------------------------------------------------------------------

llvmt = partial(ts.convert, "llvm")

def test_llvm_numeric_conversion():
    convert(ts, lts, "llvm", typenames)

def test_llvm_pointers():
    # Test pointer conversion
    for typename in typenames:
        ty = getattr(ts, tyname(typename))
        lty = getattr(lts, tyname(typename))
        assert llvmt(ts.pointer(ty)) == lts.pointer(lty)

    p = ts.pointer(ts.pointer(ts.int_))
    lp = lts.pointer(lts.pointer(lts.int_))

    # See if the conversion works
    assert llvmt(p) == lp
    # See if the consing works
    # assert llvmt(p) is lp

def test_llvm_functions():
    functype = ts.function(ts.int_, (ts.float_,))
    lfunctype = lts.function(lts.int_, (lts.float_,))
    assert llvmt(functype) == lfunctype

def test_llvm_complex():
    c1 = llvmt(ts.complex128)
    c2 = lts.struct_([('real', lts.double), ('imag', lts.double)])
    c3 = lts.struct_([('real', lts.double), ('imag', lts.double)])
    assert c1 == c2
    # assert c1 is c2
    # assert c2 is c3 # enable after upgrading llvmpy to include type hash fix

def test_llvm_object():
    assert llvmt(ts.object_) == llvm_types._pyobject_head_struct_p

def test_llvm_array():
    assert llvmt(ts.array(ts.double, 1)) == llvm_types._numpy_array
    assert llvmt(ts.array(ts.int_, 2)) == llvm_types._numpy_array
    assert llvmt(ts.array(ts.object_, 3)) == llvm_types._numpy_array

def test_llvm_range():
    assert llvmt(ts.range_) == llvm_types._pyobject_head_struct_p

#-------------------------------------------------------------------
# Numba -> ctypes
#-------------------------------------------------------------------

ct = partial(ts.convert, "ctypes")

def test_ctypes_numeric_conversion():
    convert(ts, cts, "ctypes", typenames)

def test_ctypes_pointers(): # TODO: unifiy with test_llvm_pointers
    # Test pointer conversion
    for typename in typenames:
        ty = getattr(ts, tyname(typename))
        cty = getattr(cts, tyname(typename))
        assert ct(ts.pointer(ty)) == cts.pointer(cty)

    p = ts.pointer(ts.pointer(ts.int_))
    cp = cts.pointer(cts.pointer(cts.int_))

    # See if the conversion works
    assert ct(p) == cp
    # See if the consing works
    # assert ct(p) is cp

def test_ctypes_functions(): # TODO: unifiy with test_llvm_functions
    functype = ts.function(ts.int_, (ts.float_,))
    cfunctype = cts.function(cts.int_, (cts.float_,))
    assert ct(functype) == cfunctype

def test_ctypes_complex():
    c1 = ct(ts.complex128)
    c2 = cts.struct_([('real', cts.double), ('imag', cts.double)])
    c3 = cts.struct_([('real', cts.double), ('imag', cts.double)])
    assert c1._fields_ == c2._fields_, (c1._fields_, c2._fields_)

def test_ctypes_object():
    assert ct(ts.object_) == ctypes.py_object

def test_ctypes_array():
    assert ct(ts.array(ts.double, 1)) == ctypes.py_object
    assert ct(ts.array(ts.int_, 2)) == ctypes.py_object
    assert ct(ts.array(ts.object_, 3)) == ctypes.py_object

def test_ctypes_string():
    assert ct(ts.string_) == ctypes.c_char_p
    assert ct(ts.char.pointer()) == ctypes.c_char_p

if __name__ == "__main__":
    # print(ct(ts.array(ts.double, 1)))
    for name, f in globals().items():
        if name.startswith("test_") and inspect.isfunction(f):
            f()



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

from numba.typesystem import numba_typesystem as ts

# ______________________________________________________________________
def test_mutability():
    ty = ts.function(ts.int_, (ts.float_,))
    ty.args
    try:
        ty.args = [1, 2]
    except AttributeError as e:
        pass
    else:
        raise Exception(
            "Expected: AttributeError: Cannot set attribute 'args' of type ...")
# ______________________________________________________________________

def test_pointers():
    assert ts.pointer(ts.int_) is ts.pointer(ts.int_)

def test_functions():
    functype1 = ts.function(ts.int_, (ts.float_,))
    functype2 = ts.function(ts.int_, (ts.float_,))
    functype3 = ts.function(ts.int_, (ts.float_,), is_vararg=False)
    functype4 = ts.function(ts.int_, (ts.float_,), name="hello")
    functype5 = ts.function(ts.int_, (ts.float_,), name="hello", is_vararg=False)
    functype6 = ts.function(ts.int_, (ts.float_,), name="hello", is_vararg=True)

    assert functype1 is functype2
    assert functype1 is functype3
    assert functype1 is not functype4
    assert functype1 is not functype5
    assert functype1 is not functype6

    assert functype4 is functype5
    assert functype4 is not functype6

# def test_struct():
#     s1 = ts.struct_([('a', ts.int_), ('b', ts.float_)])
#     s2 = ts.struct_([('a', ts.int_), ('b', ts.float_)])
#     assert s1 is not s2

def test_arrays():
    A = ts.array(ts.double, 1)
    B = ts.array(ts.double, 1)
    C = ts.array(ts.float_, 1)
    D = ts.array(ts.double, 2)

    assert A is B
    assert A is not C
    assert A is not D

def test_complex():
    assert ts.complex_(ts.float_) is ts.complex64
    assert ts.complex_(ts.double) is ts.complex128
    # assert ts.complex_(ts.longdouble) is ts.complex256

if __name__ == "__main__":
    test_mutability()
    test_pointers()
    test_functions()
    # test_struct()
    test_arrays()
    test_complex()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba import *
from numba.typesystem import defaults

_trace_refs_ = hasattr(sys, 'getobjects')

def define(u):
    void_star = u.pointer(u.void)
    intp_star = u.pointer(u.npy_intp)

    if _trace_refs_:
        pyobject_head_extra_fields = [
            ('ob_next', void_star),
            ('ob_prev', void_star),
        ]
    else:
        pyobject_head_extra_fields = []

    pyobject_head_fields = pyobject_head_extra_fields + [
        ('ob_refcnt', u.Py_ssize_t),
        ('ob_type', void_star),
    ]

    PyObject_HEAD = u.struct_(pyobject_head_fields, 'PyObject_HEAD')
    PyArray = u.struct_(pyobject_head_fields + [
        ("data", void_star),
        ("nd", u.int32),
        ("dimensions", intp_star),
        ("strides", intp_star),
        ("base", void_star),
        ("descr", void_star),
        ("flags", u.int32),
        ("weakreflist", void_star),
      ])

    PyCFunctionObject = u.struct_([
        ('head', PyObject_HEAD),
        ('m_ml', void_star),
        ('m_self', u.object_),
        ('m_module', u.object_),
    ])

    # TODO: Parse C and Cython header files...
    NumbaFunctionObject = u.struct_([
        ('pycfunction',         PyCFunctionObject),
        ('flags',               u.int_),
        ('func_dict',           u.object_),
        ('func_weakreflist',    u.object_),
        ('func_name',           u.object_),
        ('func_doc',            u.object_),
        ('func_code',           u.object_),
        ('func_closure',        u.object_),
    ])

    return locals()

globals().update(define(defaults.numba_typesystem))


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import unittest
from numba import autojit

# ______________________________________________________________________
# Helpers

class SpecialException(Exception):
    pass

# ______________________________________________________________________

@autojit
def raise1():
    raise SpecialException

@autojit
def raise2():
    raise SpecialException("hello")

# ______________________________________________________________________

class TestRaise(unittest.TestCase):

    def _assert_raises(self, func, expected_args):
        try:
            func()
        except SpecialException as e:
            assert e.args == tuple(expected_args), (e.args, expected_args)
        else:
            raise AssertionError("Expected exception")

    def test_raise(self):
        self._assert_raises(raise1, [])
        self._assert_raises(raise2, ["hello"])

if __name__ == "__main__":
    unittest.main()


import ast

class FixMissingLocations(ast.NodeVisitor):
    """
    Fix missing source position information.
    """

    def __init__(self, lineno, col_offset, override=False):
        self.lineno = lineno
        self.col_offset = col_offset
        self.override = override

    def visit(self, node):
        super(FixMissingLocations, self).visit(node)
        if not hasattr(node, 'lineno') or self.override:
            node.lineno = self.lineno
            node.col_offset = self.col_offset
        else:
            self.lineno = node.lineno
            self.col_offset = node.col_offset


# -*- coding: utf-8 -*-

"""
Extension method types.
"""

from __future__ import print_function, division, absolute_import
from numba.typesystem import types, numbatypes

#------------------------------------------------------------------------
# Extension Method Types
#------------------------------------------------------------------------

class ExtMethodType(types.function):
    typename = "extmethod"
    argnames = ["return_type", "args", ("name", None), ("is_vararg", False),
                ("is_class_method", False), ("is_static_method", False)]
    flags = ["function", "object"]

    @property
    def is_bound_method(self):
        return not (self.is_class_method or self.is_static_method)

class AutojitMethodType(types.NumbaType):
    typename = "autojit_extmethod"
    flags = ["object"]

#------------------------------------------------------------------------
# Method Signature Comparison
#------------------------------------------------------------------------

def drop_self(type):
    if type.is_static_method or type.is_class_method:
        return type.args

    assert len(type.args) >= 1 and type.args[0].is_extension, type
    return type.args[1:]

def equal_signature_args(t1, t2):
    """
    Compare method signatures without regarding the 'self' type (which is
    set to the base extension type in the base class, and the derived
    extension type in the derived class).
    """
    return (t1.is_static_method == t2.is_static_method and
            t1.is_class_method == t2.is_class_method and
            t1.is_bound_method == t2.is_bound_method and
            drop_self(t1) == drop_self(t2))

def equal_signatures(t1, t2):
    return (equal_signature_args(t1, t2) and
            t1.return_type == t2.return_type)

def extmethod_to_function(ty):
    return numbatypes.function(ty.return_type, ty.args, ty.name)


"""
Implements threads using llvm cbuilder. Taken from
numbapro/vectorizers/parallel.py
"""

from llvm.core import *
from llvm_cbuilder import *
import llvm_cbuilder.shortnames as C

import sys

class PThreadAPI(CExternal):
    '''external declaration of pthread API
    '''
    pthread_t = C.void_p

    pthread_create = Type.function(C.int,
                                   [C.pointer(pthread_t),  # thread_t
                                    C.void_p,              # thread attr
                                    C.void_p,              # function
                                    C.void_p])             # arg

    pthread_join = Type.function(C.int, [C.void_p, C.void_p])


class WinThreadAPI(CExternal):
    '''external declaration of pthread API
    '''
    _calling_convention_ = CC_X86_STDCALL

    handle_t = C.void_p

    # lpStartAddress is an LPTHREAD_START_ROUTINE, with the form
    # DWORD ThreadProc (LPVOID lpdwThreadParam )
    CreateThread = Type.function(handle_t,
                                   [C.void_p,            # lpThreadAttributes (NULL for default)
                                    C.intp,              # dwStackSize (0 for default)
                                    C.void_p,            # lpStartAddress
                                    C.void_p,            # lpParameter
                                    C.int32,             # dwCreationFlags (0 for default)
                                    C.pointer(C.int32)]) # lpThreadId (NULL if not required)

    # Return is WAIT_OBJECT_0 (0x00000000) to indicate the thread exited,
    # or WAIT_ABANDONED, WAIT_TIMEOUT, WAIT_FAILED for other conditions.
    WaitForSingleObject = Type.function(C.int32,
                                    [handle_t, # hHandle
                                     C.int32])   # dwMilliseconds (INFINITE == 0xFFFFFFFF means wait forever)

    CloseHandle = Type.function(C.int32, [handle_t])


class ParallelUFuncPosixMixin(object):
    '''ParallelUFunc mixin that implements _dispatch_worker to use pthread.
    '''
    def _dispatch_worker(self, worker, contexts, num_thread):
        api = PThreadAPI(self)
        NULL = self.constant_null(C.void_p)

        threads = self.array(api.pthread_t, num_thread, name='threads')

        # self.debug("launch threads")

        with self.for_range(num_thread) as (loop, i):
            status = api.pthread_create(threads[i].reference(), NULL, worker,
                                        contexts[i].reference().cast(C.void_p))
            with self.ifelse(status != self.constant_null(status.type)) as ifelse:
                with ifelse.then():
                    # self.debug("Error at pthread_create: ", status)
                    self.unreachable()

        with self.for_range(num_thread) as (loop, i):
            status = api.pthread_join(threads[i], NULL)
            with self.ifelse(status != self.constant_null(status.type)) as ifelse:
                with ifelse.then():
                    # self.debug("Error at pthread_join: ", status)
                    self.unreachable()


class ParallelUFuncWindowsMixin(object):
    '''ParallelUFunc mixin that implements _dispatch_worker to use Windows threading.
    '''
    def _dispatch_worker(self, worker, contexts, num_thread):
        api = WinThreadAPI(self)
        NULL = self.constant_null(C.void_p)
        lpdword_NULL = self.constant_null(C.pointer(C.int32))
        zero = self.constant(C.int32, 0)
        intp_zero = self.constant(C.intp, 0)
        INFINITE = self.constant(C.int32, 0xFFFFFFFF)

        threads = self.array(api.handle_t, num_thread, name='threads')

        # self.debug("launch threads")
        # TODO error handling

        with self.for_range(num_thread) as (loop, i):
            threads[i] = api.CreateThread(NULL, intp_zero, worker,
                               contexts[i].reference().cast(C.void_p),
                               zero, lpdword_NULL)

        with self.for_range(num_thread) as (loop, i):
            api.WaitForSingleObject(threads[i], INFINITE)
            api.CloseHandle(threads[i])


if sys.platform == 'win32':
    ParallelMixin = ParallelUFuncWindowsMixin
else:
    ParallelMixin = ParallelUFuncPosixMixin


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import warnings
from functools import partial

from numba.config import config

try:
    import pygments
except ImportError as e:
    pygments = None
else:
    from pygments import highlight
    from pygments.lexers import PythonLexer, LlvmLexer
    from pygments.formatters import HtmlFormatter, TerminalFormatter

# ______________________________________________________________________

if pygments:

    lexers = {
        "python": PythonLexer,
        "llvm": LlvmLexer,
    }

    formatters = {
        "html": HtmlFormatter,
        "console": partial(TerminalFormatter, bg=config.terminal_background),
    }

    def lex_source(code, lexer="python", output='html', inline_css=True):
        """
        >>> lex_source("print 'hello world'", "python", "html")
        <div ...> ... </div>
        """
        if not config.colour:
            return code

        Lexer = lexers[lexer]
        Formatter = formatters[output]
        result = highlight(code, Lexer(), Formatter(noclasses=inline_css))
        return result.rstrip()

else:
    def lex_source(code, *args, **kwargs):
        warnings.warn("Pygments not installed")
        return code


"""
Thanks to Aron Ahmadia
"""

from __future__ import division, print_function

import sys
import math
import numba
from numba import jit, autojit, size_t
import numpy as np
import numpy.testing as npt
try:
    from skimage import img_as_float
except ImportError as e:
    print("skimage not available, skipping")
    sys.exit()

SCALAR_DTYPE = np.float64
# This doesn't work :(
# SCALAR_TYPE  = numba.typeof(SCALAR_DTYPE)
SCALAR_TYPE = numba.float64


def window_floor(idx, radius):
    if radius > idx:
        return 0
    else:
        return idx - radius


def window_ceil(idx, ceil, radius):
    if idx + radius > ceil:
        return ceil
    else:
        return idx + radius


def distance(image, r0, c0, r1, c1):
    d = image[r0, c0, 0] - image[r1, c1, 0]
    s = d * d
    for i in range(1, 3):
        d = image[r0, c0, i] - image[r1, c1, i]
        s += d * d
    return math.sqrt(s)


def pixel_distance(pixel1, pixel2):
    d = pixel1[0] - pixel2[0]
    s = d*d
    for i in range(1, 3):
        d = pixel1[i] - pixel2[i]
        s += d*d
    return math.sqrt(s)


def np_distance(pixel1, pixel2):
    return np.linalg.norm(pixel1-pixel2, 2)


sqrt_3 = math.sqrt(3.0)


def g(d):
    return 1.0 - d/sqrt_3


def np_g(x, y):
    return 1.0 - np_distance(x, y)/sqrt_3


def kernel(image, state, state_next, window_radius):
    changes = 0

    height = image.shape[0]
    width = image.shape[1]

    for j in xrange(width):
        for i in xrange(height):

            winning_colony = state[i, j, 0]
            defense_strength = state[i, j, 1]

            for jj in xrange(window_floor(j, window_radius),
                             window_ceil(j+1, width, window_radius)):
                for ii in xrange(window_floor(i, window_radius),
                                 window_ceil(i+1, height, window_radius)):
                    if (ii == i and jj == j):
                        continue

                    d = image[i, j, 0] - image[ii, jj, 0]
                    s = d * d
                    for k in range(1, 3):
                        d = image[i, j, k] - image[ii, jj, k]
                        s += d * d
                    gval = 1.0 - math.sqrt(s)/sqrt_3

                    attack_strength = gval * state[ii, jj, 1]

                    if attack_strength > defense_strength:
                        defense_strength = attack_strength
                        winning_colony = state[ii, jj, 0]
                        changes += 1

            state_next[i, j, 0] = winning_colony
            state_next[i, j, 1] = defense_strength

    return changes


def growcut(image, state, max_iter=20, window_size=3):
    """Grow-cut segmentation (Numba accelerated).

    Parameters
    ----------
    image : (M, N) ndarray
        Input image.
    state : (M, N, 2) ndarray
        Initial state, which stores (foreground/background, strength) for
        each pixel position or automaton.  The strength represents the
        certainty of the state (e.g., 1 is a hard seed value that remains
        constant throughout segmentation).
    max_iter : int, optional
        The maximum number of automata iterations to allow.  The segmentation
        may complete earlier if the state no longer varies.
    window_size : int, optional
        Size of the neighborhood window.

    Returns
    -------
    mask : ndarray
        Segmented image.  A value of zero indicates background, one foreground.

    """

    image = img_as_float(image)

    window_radius = (window_size - 1) // 2

    changes = 1
    n = 0

    state_next = np.empty_like(state)

    while changes > 0 and n < max_iter:
        changes = 0
        n += 1
        changes = kernel(image, state, state_next, window_radius)
        state_next, state = state, state_next
        #print n, changes
        print('.', end='')
    print('')
    return state_next[:, :, 0]


def create_numba_funcs(scalar_type=SCALAR_TYPE):
    this = sys.modules[__name__]

    pixel_type = scalar_type[:]
    image_type = scalar_type[:, :, :]
    state_type = scalar_type[:, :, :]

    this._numba_window_floor = jit(nopython=True,
                                   argtypes=[size_t, size_t],
                                   restype=size_t)(_py_window_floor)

    this._numba_window_ceil = jit(nopython=True,
                                  argtypes=[size_t, size_t, size_t],
                                  restype=size_t)(_py_window_ceil)

    this._numba_distance = jit(nopython=True,
                               argtypes=[image_type,
                                         size_t, size_t, size_t, size_t],
                               restype=scalar_type)(_py_distance)

    this._numba_np_distance = jit(nopython=False,
                                  argtypes=[pixel_type, pixel_type],
                                  restype=scalar_type)(_py_np_distance)

    this._numba_g = jit(nopython=True,
                        argtypes=[scalar_type],
                        restype=scalar_type)(_py_g)

    this._numba_np_g = jit(nopython=False,
                           argtypes=[pixel_type, pixel_type],
                           restype=scalar_type)(_py_np_g)

    this._numba_kernel = autojit(nopython=True)(_py_kernel)
    # the below code does not work
    # this._numba_kernel        = jit(nopython=False,
    #                                  argtypes=[image_type,
    #                                            state_type,
    #                                            state_type,
    #                                            size_t],
    #                                  restype=int_,
    #                                  attack_strength=scalar_type,
    #                                  defense_strength=scalar_type,
    #                                  winning_colony=scalar_type)(_py_kernel)


def debug():
    this = sys.modules[__name__]
    this.window_floor = _py_window_floor
    this.window_ceil = _py_window_ceil
    this.distance = _py_distance
    this.np_distance = _py_np_distance
    this.g = _py_g
    this.np_g = _py_np_g
    this.kernel = _py_kernel


def optimize():
    this = sys.modules[__name__]
    this.window_floor = _numba_window_floor
    this.window_ceil = _numba_window_ceil
    this.distance = _numba_distance
    this.np_distance = _numba_np_distance
    this.g = _numba_g
    this.np_g = _numba_np_g
    this.kernel = _numba_kernel


# protected Pythonic versions of code:
_py_window_floor = window_floor
_py_window_ceil = window_ceil
_py_distance = distance
_py_np_distance = np_distance
_py_g = g
_py_np_g = np_g
_py_kernel = kernel


def test_window_floor_ceil():

    assert 3 == window_floor(4, 1)
    assert 0 == window_floor(1, 4)

    assert 3 == window_ceil(3, 3, 1)
    assert 5 == window_ceil(4, 5, 1)


def test_distance():
    image = np.zeros((2, 2, 3), dtype=SCALAR_DTYPE)
    image[0, 1] = [1, 1, 1]
    image[1, 0] = [0.5, 0.5, 0.5]

    assert 0.0 == distance(image, 0, 0, 0, 0)
    assert abs(math.sqrt(3) - distance(image, 0, 0, 0, 1)) < 1e-15
    assert abs(math.sqrt(3/4) - distance(image, 0, 1, 1, 0)) < 1e-15

    pixel1 = np.asarray([0.0, 0.0, 0.0], dtype=SCALAR_DTYPE)
    pixel2 = np.asarray([1.0, 1.0, 1.0], dtype=SCALAR_DTYPE)
    pixel3 = np.asarray([0.5, 0.5, 0.5], dtype=SCALAR_DTYPE)

    assert 0.0 == np_distance(pixel1, pixel1)
    assert abs(math.sqrt(3) - np_distance(pixel1, pixel2)) < 1e-15
    assert abs(math.sqrt(3/4) - np_distance(pixel2, pixel3)) < 1e-15


def test_g():
    image = np.zeros((2, 2, 3), dtype=SCALAR_DTYPE)
    image[0, 1] = [1, 1, 1]
    image[1, 0] = [0.5, 0.5, 0.5]

    assert 1.0 == g(distance(image, 0, 0, 0, 0))
    assert abs(0 - g(distance(image, 0, 0, 0, 1))) < 1e-15
    assert abs(0.5 - g(distance(image, 0, 1, 1, 0))) < 1e-15

    pixel1 = np.asarray([0.0, 0.0, 0.0], dtype=SCALAR_DTYPE)
    pixel2 = np.asarray([1.0, 1.0, 1.0], dtype=SCALAR_DTYPE)
    pixel3 = np.asarray([0.5, 0.5, 0.5], dtype=SCALAR_DTYPE)

    assert 1.0 == np_g(pixel1, pixel1)
    assert abs(0 - np_g(pixel1, pixel2)) < 1e-15
    assert abs(0.5 - np_g(pixel2, pixel3)) < 1e-15


def test_kernel():
    image = np.zeros((3, 3, 3), dtype=SCALAR_DTYPE)
    state = np.zeros((3, 3, 2), dtype=SCALAR_DTYPE)
    state_next = np.empty_like(state)

    # colony 1 is strength 1 at position 0,0
    # colony 0 is strength 0 at all other positions
    state[0, 0, 0] = 1
    state[0, 0, 1] = 1

    # window_size 1, colony 1 should propagate to three neighbors
    changes = kernel(image, state, state_next, 1)
    assert(3 == changes)
    npt.assert_array_equal(state_next[0:2, 0:2], 1)
    npt.assert_array_equal(state_next[2, :], 0)
    npt.assert_array_equal(state_next[2, :], 0)

    # window_size 1, colony 1 should propagate to entire image
    changes = kernel(image, state, state_next, 2)
    assert(8 == changes)
    npt.assert_array_equal(state_next, 1)


def test():
    test_window_floor_ceil()
    test_distance()
    test_g()
    test_kernel()


# create numba versions of code
create_numba_funcs()


if __name__ == "__main__":
    # always verify pure Python code first
    test()
    # then test optimized variants
    optimize()
    test()


# replace default function calls with numba calls
optimize()


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import numba
from numba import autojit, double

import unittest
# from unittest import FunctionTestCase as testcase

import numpy as np

tests = []
def testcase(f):
    tests.append(f)
    return f

#----------------------------------------------------------------------------
# Simple isolated tests
#----------------------------------------------------------------------------

@testcase
def test_simple_prange_shared():
    @autojit(warn=False)
    def simple_prange_shared():
        """
        >>> simple_prange_shared()
        20L
        """
        result = np.empty(1, dtype=np.int64)
        shared = 20

        for i in numba.prange(1):
            result[0] = shared
        return result[0]
    assert simple_prange_shared() == 20

@testcase
def test_simple_prange_private():
    @autojit(warn=False)
    def simple_prange_private():
        """
        >>> simple_prange_private()
        10L
        """
        result = np.empty(1, dtype=np.int64)
        var = 20
        for i in numba.prange(1):
            var = 10
            result[0] = var
        return result[0]
    assert simple_prange_private() == 10

@testcase
def test_simple_prange_lastprivate():
    @autojit(warn=False)
    def simple_prange_lastprivate():
        """
        >>> simple_prange_lastprivate()
        10
        """
        var = 20
        for i in numba.prange(1):
            var = 10
        return var
    assert simple_prange_lastprivate() == 10
    

@testcase
def test_simple_prange_reduction():
    @autojit(warn=False)
    def simple_prange_reduction():
        """
        >>> simple_prange_reduction()
        15
        """
        var = 10
        for i in numba.prange(1):
            var += 5
        return var
    assert simple_prange_reduction() == 15

#----------------------------------------------------------------------------
# Error Tests
#----------------------------------------------------------------------------

@autojit(warn=False)
def prange_reduction_error():
    """
    DISABLED.

    >> prange_reduction_error()
    Traceback (most recent call last):
        ...
    NumbaError: 32:8: Local variable  'sum' is not bound yet
    """
    for i in numba.prange(10):
        sum += i
    sum = 0.0
    return sum

#----------------------------------------------------------------------------
# Advanced Tests
#----------------------------------------------------------------------------

@testcase
def test_prange_reduction2():
    @autojit(warn=False)
    def prange_reduction2():
        """
        >>> prange_reduction2()
        49999995000000.0
        """
        sum = 0.0
        for i in numba.prange(10000000):
            sum += i
        return sum
    assert prange_reduction2() == 49999995000000.0

@testcase
def test_prange_reduction_and_privates():
    @autojit(warn=False)
    def prange_reduction_and_privates():
        """
        >>> prange_reduction_and_privates()
        100.0
        """
        sum = 10.0
        for i in numba.prange(10):
            j = i * 2
            sum += j
        return sum
    assert prange_reduction_and_privates() == 100.0

@testcase
def test_prange_lastprivate():
    @autojit(warn=False)
    def prange_lastprivate():
        """
        >>> prange_lastprivate()
        100.0
        18
        """
        sum = 10.0
        for i in numba.prange(10):
            j = i * 2
            sum += j
        print(sum)
        return j
    assert prange_lastprivate() == 18

@testcase
def test_prange_shared_privates_reductions():
    @autojit(warn=False)
    def prange_shared_privates_reductions(shared):
        """
        >>> prange_shared_privates_reductions(2.0)
        100.0
        """
        sum = 10.0

        for i in numba.prange(10):
            j = i * shared
            sum += j
        shared = 3.0
        return sum
    assert prange_shared_privates_reductions(2.0) == 100.0

@testcase
def test_test_sum2d():
    @autojit(warn=False)
    def test_sum2d(A):
        """
        >>> a = np.arange(100).reshape(10, 10)
        >>> test_sum2d(a)
        4950.0
        >>> test_sum2d(a.astype(np.complex128))
        (4950+0j)
        >>> np.sum(a)
        4950
        """
        sum = 0.0
        for i in numba.prange(A.shape[0]):
            for j in range(A.shape[1]):
                # print(i, j)
                sum += A[i, j]
        return sum

    a = np.arange(100).reshape(10, 10)
    assert test_sum2d(a) == 4950.0
    assert test_sum2d(a.astype(np.complex128)) == 4950+0j
    assert np.sum(a) == 4950

@testcase
def test_test_prange_in_closure():
    @autojit(warn=False)
    def test_prange_in_closure(x):
        """
        >>> test_prange_in_closure(2.0)()
        1000.0
        """
        sum = 10.0
        N = 10

        @double()
        def inner():
            sum = 100.0
            for i in numba.prange(N):
                for j in range(N):
                    sum += i * x
            return sum
        return inner
    assert test_prange_in_closure(2.0)() == 1000.0

@testcase
def test_test_prange_in_closure2():
    @autojit(warn=False)
    def test_prange_in_closure2(x):
        """
        >>> test_prange_in_closure2(2.0)()
        10000.0
        """
        sum = 10.0
        N = 10

        @double()
        def inner():
            sum = 100.0
            for i in numba.prange(N):
                for j in range(N):
                    sum += (i * N + j) * x
            return sum
        return inner

    assert test_prange_in_closure2(2.0)() == 10000.0

if __name__ == '__main__':
    # unittest.main()
    for test in tests:
        print("running", test.__name__)
        test()


# -*- coding: utf-8 -*-

"""
Initial AST validation and normalization.
"""

from __future__ import print_function, division, absolute_import

import ast
from numba import error, nodes

class ValidateAST(ast.NodeVisitor):
    "Validate AST"

    #------------------------------------------------------------------------
    # Validation
    #------------------------------------------------------------------------

    def visit_GeneratorExp(self, node):
        raise error.NumbaError(
                node, "Generator comprehensions are not yet supported")

    def visit_SetComp(self, node):
        raise error.NumbaError(
                node, "Set comprehensions are not yet supported")

    def visit_DictComp(self, node):
        raise error.NumbaError(
                node, "Dict comprehensions are not yet supported")

    # def visit_Raise(self, node):
    #     if node.tback:
    #         raise error.NumbaError(
    #             node, "Traceback argument to raise not supported")

    def visit_For(self, node):
        if not isinstance(node.target, (ast.Name, ast.Attribute,
                                        nodes.TempStoreNode)):
            raise error.NumbaError(
                node.target, "Only a single target iteration variable is "
                             "supported at the moment")

        self.generic_visit(node)

    def visit_With(self, node):
        self.visit(node.context_expr)
        if node.optional_vars:
            raise error.NumbaError(
                node.context_expr,
                "Only 'with python' and 'with nopython' is "
                "supported at this moment")

        self.generic_visit(node)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import ast

from numba import error
from numba.ndarray_helpers import NumpyArray

is_object = lambda t: t.is_object and not t.is_array


class ArrayValidator(ast.NodeVisitor):
    """
    Validate array usage, depending on array representation
    (i.e. numpy vs. LLArray)
    """

    def __init__(self, env):
        self.env = env
        self.foreign = not issubclass(env.crnt.array, NumpyArray)

    def visit_CoercionNode(self, node):
        t1, t2 = node.type, node.node.type
        if self.foreign and t1.is_array ^ t2.is_array:
            raise error.NumbaError(node, "Cannot coerce non-numpy array %s" %
                                         self.env.crnt.array)

    visit_CoerceToObject = visit_CoercionNode
    visit_CoerceToNative = visit_CoercionNode

    def visit_MultiArrayAPINode(self, node):
        if self.foreign:
            signature = node.signature
            types = (signature.return_type,) + signature.argtypes
            for ty in types:
                if not ty.is_array:
                    raise TypeError("Cannot pass array %s as NumPy array" %
                                                        self.env.crnt.array)

    # TODO: Calling other numba functions with different array
    # TODO: representations may corrupt things


# -*- coding: UTF-8 -*-

"""
HTML annotation rendering. Heavily based on Cython/Compiler/Annotate.py
"""

from __future__ import print_function, division, absolute_import

import sys
import cgi
import os
import re
from .annotate import format_annotations, groupdict, A_c_api
from .step import Template

try:
    from pygments import highlight
    from pygments.lexers import PythonLexer
    from pygments.lexers import LlvmLexer
    from pygments.formatters import HtmlFormatter
    pygments_installed = True
except ImportError:
    pygments_installed = False

def render(annotation_blocks, emit=sys.stdout.write,
           intermediate_names=(), inline=True):
    """
    Render a Program as html.
    """
    root = os.path.join(os.path.dirname(__file__))
    if inline:
        templatefile = os.path.join(root, 'annotate_inline_template.html')
    else:
        templatefile = os.path.join(root, 'annotate_template.html')

    with open(templatefile, 'r') as f:
        template = f.read()

    py_c_api = re.compile(u'(Py[A-Z][a-z]+_[A-Z][a-z][A-Za-z_]+)\(')

    data = {'blocks': []}

    for i, block in enumerate(annotation_blocks):
        python_source = block['python_source']
        intermediates = block['intermediates']
        data['blocks'].append({'lines':[]})

        for num, source in sorted(python_source.linemap.items()):

            types = {}
            if num in python_source.annotations.keys():
                for a in python_source.annotations[num]:
                    if a.type == 'Types':
                        name = a.value[0]
                        type = a.value[1]
                        types[name] = type
            
            types_str = ','.join(name + ':' + type for name, type in types.items())

            python_calls = 0
            llvm_nums = intermediates[0].linenomap[num]
            llvm_ir = ''
            for llvm_num in llvm_nums:
                ir = intermediates[0].source.linemap[llvm_num]
                if re.search(py_c_api, ir):
                    python_calls += 1

                if pygments_installed:

                    class LlvmHtmlFormatter(HtmlFormatter):
                        def wrap(self, source, outfile):
                            return self._wrap_code(source)
                        def _wrap_code(self, source):
                            for i, t in source:
                                yield i, t

                    ir = highlight(ir, LlvmLexer(), LlvmHtmlFormatter())
                       
                llvm_ir += '<div>' + ir + '</div>'

            if python_calls > 0:
                tag = '*'
                tag_css = 'tag'
            else:
                tag = ''
                tag_css = ''

            if num == python_source.linemap.keys()[0]:
                firstlastline = 'firstline'
            elif num == python_source.linemap.keys()[-1]:
                firstlastline = 'lastline'
            else:
                firstlastline = 'innerline'

            if pygments_installed:

                class PythonHtmlFormatter(HtmlFormatter):
                    def wrap(self, source, outfile):
                        return self._wrap_code(source)
                    def _wrap_code(self, source):
                        for i, t in source:
                            yield i, t

                source = highlight(source, PythonLexer(), PythonHtmlFormatter())
               
            data['blocks'][-1]['func_call'] = block['func_call']
            data['blocks'][-1]['func_call_filename'] = block['func_call_filename']
            data['blocks'][-1]['func_call_lineno'] = block['func_call_lineno']
            data['blocks'][-1]['lines'].append({'id': str(i) + str(num),
                                  'num':str(num) + tag,
                                  'tag':tag_css,
                                  'python_source':source,
                                  'llvm_source':llvm_ir,
                                  'types':types_str,
                                  'firstlastline':firstlastline})

    css_theme_file = os.path.join(root, 'jquery-ui.min.css')
    with open(css_theme_file, 'r') as f:
        css_theme = f.read()
    data['jquery_theme'] = css_theme

    jquery_lib_file = os.path.join(root, 'jquery.min.js')
    with open(jquery_lib_file, 'r') as f:
        jquery_lib = f.read()
    data['jquery_lib'] = jquery_lib

    jquery_ui_lib_file = os.path.join(root, 'jquery-ui.min.js')
    with open(jquery_ui_lib_file, 'r') as f:
        jquery_ui_lib = f.read()
    data['jquery_ui_lib'] = jquery_ui_lib

    html = Template(template).expand(data)

    emit(html)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import os

from .. import generator

root = os.path.dirname(os.path.abspath(__file__))

def schema_filename(name):
    return os.path.join(root, name)

def load_schema(name):
    schema_fn = schema_filename(name)
    schema_name = os.path.basename(schema_fn)
    schema_str = open(schema_fn).read()
    return schema_name, schema_str

def generate_in_memory(schema_name, codegens):
    file_allocator = generator.MemoryFileAllocator()
    schema_name, schema_str = load_schema(schema_name)
    generator.generate(schema_name, schema_str, codegens, file_allocator)
    return file_allocator



# -*- coding: utf-8 -*-

"""
Generate Python visitors and Cython pxd files.
"""

from __future__ import print_function, division, absolute_import

import os

from . import generator
from . import naming
from .formatting import py_formatter

#------------------------------------------------------------------------
# Code Formatting
#------------------------------------------------------------------------

interface_class = '''

def iter_fields(node):
    """
    Yield a tuple of ``(fieldname, value)`` for each field in ``node._fields``
    that is present on *node*.
    """
    result = []
    for field in node._fields:
        try:
            result.append((field, getattr(node, field)))
        except AttributeError:
            pass

    return result

class GenericVisitor(object):

    def visit(self, node):
        return node.accept(self)

    def generic_visit(self, node):
        """Called explicitly by the user from an overridden visitor method"""
        raise NotImplementedError

'''

pxd_interface_class = """\
cimport %s

cdef class GenericVisitor(object):
    cpdef generic_visit(self, node)
""" % (naming.nodes,)


# TODO: We can also make 'visitchildren' dispatch quickly

visitor_class = '''
from %s import GenericVisitor, iter_fields
from %s import AST

__all__ = ['Visitor']

class Visitor(GenericVisitor):

    def generic_visit(self, node):
        for field, value in iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, AST):
                        item.accept(self)
            elif isinstance(value, AST):
                value.accept(self)

''' % (naming.interface, naming.nodes)

transformer_class = """
from %s import GenericVisitor, iter_fields
from %s import AST

__all__ = ['Transformer']

class Transformer(GenericVisitor):

    def generic_visit(self, node):
        for field, old_value in iter_fields(node):
            old_value = getattr(node, field, None)
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, AST):
                        value = value.accept(self)
                        if value is None:
                            continue
                        elif not isinstance(value, AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, AST):
                new_node = old_value.accept(self)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node

""" % (naming.interface, naming.nodes,)

pxd_visitor_class = """
from %s cimport GenericVisitor

cdef class Visitor(GenericVisitor):
    pass
""" % (naming.interface,)

pxd_transformer_class = """
from %s cimport GenericVisitor

cdef class Transformer(GenericVisitor):
    pass
""" % (naming.interface,)

#------------------------------------------------------------------------
# Code Formatting
#------------------------------------------------------------------------

def make_visit_stats(schema, fields, inplace):
    stats = []

    # ["self.attr_x"]
    field_accessors = py_formatter.get_fields(fields, obj="node")

    for field, field_access in zip(fields, field_accessors):
        field_type = str(field.type)
        if field_type not in schema.dfns and field_type not in schema.types:
            # Not an AST node
            continue

        s = "%s.accept(self)" % field_access

        if inplace:
            # Mutate in-place (transform)
            s = "%s = %s" % (field_access, s)

        if field.opt:
            # Guard for None
            s = "if %s is not None: %s" % (field_access, s)

        stats.append(s)

    if inplace:
        stats.append("return node")

    return stats or ["pass"]

#------------------------------------------------------------------------
# Method Generation
#------------------------------------------------------------------------

class Method(object):
    def __init__(self, schema, name, fields):
        self.schema = schema
        self.name = name
        self.fields = fields

class InterfaceMethod(Method):
    def __str__(self):
        return (
           "    def visit_%s(self, node):\n"
           "        raise NotImplementedError\n"
           "\n"
       ) % (self.name,)

class PyMethod(Method):

    inplace = None

    def __str__(self):
        stats = make_visit_stats(self.schema, self.fields, self.inplace)
        return (
           "    def visit_%s(self, node):\n"
           "        %s\n"
           "\n"
       ) % (self.name, py_formatter.format_stats("\n", 8, stats))

class PyVisitMethod(PyMethod):
    inplace = False

class PyTransformMethod(PyVisitMethod):
    inplace = True

class PxdMethod(Method):
    def __str__(self):
        return "    cpdef visit_%s(self, %s.%s node)\n" % (self.name,
                                                           naming.nodes,
                                                           self.name)

#------------------------------------------------------------------------
# Code Generators
#------------------------------------------------------------------------

class VisitorCodegen(generator.SimpleCodegen):
    """
    Generate Python AST nodes.
    """

    def __init__(self, out_filename, preamble, Method):
        super(VisitorCodegen, self).__init__(out_filename)
        self.preamble = preamble
        self.Method = Method

    def emit_preamble(self, emitter, schema):
        emitter.emit(self.preamble)

    def emit_sum(self, emitter, schema, rulename, rule, sumtype):
        fields = schema.types[sumtype]
        emitter.emit(self.Method(schema, sumtype, fields))

#------------------------------------------------------------------------
# Global Exports
#------------------------------------------------------------------------

codegens = [
    VisitorCodegen(naming.interface + '.py', interface_class, InterfaceMethod),
    VisitorCodegen(naming.interface + '.pxd', pxd_interface_class, PxdMethod),
    VisitorCodegen(naming.visitor + '.py', visitor_class, PyVisitMethod),
    generator.UtilityCodegen(naming.visitor + '.pxd', pxd_visitor_class),
    VisitorCodegen(naming.transformer + '.py', transformer_class, PyTransformMethod),
    generator.UtilityCodegen(naming.transformer + '.pxd', pxd_transformer_class),
]

if __name__ == '__main__':
    root = os.path.dirname(os.path.abspath(__file__))
    testdir = os.path.join(root, "tests")
    schema_filename = os.path.join(testdir, "testschema1.asdl")
    generator.generate_from_file(schema_filename, codegens, root)


# -*- coding: utf-8 -*-

"""
ctypes type universe.
"""

from __future__ import print_function, division, absolute_import

import ctypes

from numba.typesystem.itypesystem import consing, tyname
from numba.typesystem import universe
from numba.typesystem import numbatypes as ts

domain_name = "ctypes"

# ______________________________________________________________________

nb2ctypes = {
    ts.float32:    ctypes.c_float,
    ts.float64:    ctypes.c_double,
    # ts.float128:   ctypes.c_longdouble,
    ts.object_:    ctypes.py_object,
    ts.void:       None,
    ts.string_:    ctypes.c_char_p,
}

def cint(name):
    ty = getattr(ts, name)
    cname = "c_int" if ty.signed else "c_uint"
    nb2ctypes[ty] = getattr(ctypes, cname + str(ty.itemsize * 8))

for name in map(tyname, universe.int_typenames):
    cint(name)

globals().update((tyname(ty.typename), cty)
                     for ty, cty in nb2ctypes.iteritems())
# float_, double, longdouble = float32, float64, float128
float_, double = float32, float64


ctypes_map = dict((cty, ty) for ty, cty in nb2ctypes.iteritems())

if ctypes.c_double == ctypes.c_longdouble:
    # Ctypes and numpy can disagree on longdouble. If ctypes assume
    # double == longdouble, make sure we don't assume longdouble, and then
    # later use the numpy representation
    ctypes_map[ctypes.c_double] = ts.float64

# ______________________________________________________________________

@consing
def struct_(fields, name=None, readonly=False, packed=False):
    class Struct(ctypes.Structure):
        _fields_ = fields
        if packed:
            _pack_ = 1

    return Struct

@consing
def function(rettype, argtypes, name=None, is_vararg=False):
    assert not is_vararg
    return ctypes.CFUNCTYPE(rettype, *argtypes)

@consing
def pointer(base_type):
    if base_type in (ctypes.c_char, ctypes.c_byte):
        return string_
    return  ctypes.POINTER(base_type)

carray  = consing(lambda base_type, size: base_type * size)


from __future__ import print_function, division, absolute_import
import os
import contextlib

from numba import jit, autojit
from numba.tests.issues import issue_305_helper1, issue_305_helper2

try:
    from imp import reload
except ImportError:
    pass

# Thanks to @bfredl

env = { '__file__': __file__, '__name__': __name__ ,
        'jit': jit, 'autojit': autojit }

root = os.path.dirname(os.path.abspath(__file__))
fn1 = os.path.join(root, 'issue_305_helper1.py')
fn2 = os.path.join(root, 'issue_305_helper2.py')

new_source1 = """
from numba import jit, autojit

@jit('i8()')
def test():
    return 1
"""

new_source2 = """
from numba import jit, autojit

@autojit
def test2():
    return 1
"""

@contextlib.contextmanager
def newsource(fn, mod, reexec=True):
    old_source = open(fn).read()
    try:
        open(fn, 'w').write(new_source1)
        if reexec: reload(mod)
        yield
    finally:
        open(fn, 'w').write(old_source)


def test_fetch_latest_source():
    """
    When reloading new versions of the same module into the same session (i.e.
    an interactive ipython session), numba sometimes gets the wrong version of
    the source from inspect.getsource()
    """
    with newsource(fn1, issue_305_helper1):
        assert issue_305_helper1.test() == issue_305_helper1.test.py_func()

def test_no_auto_reload():
    """
    In this case autojit 'sees' the new version of the source even if it
    hasn't been reloaded. This could be fixed by fetching the ast directly
    at declaration time rather that at first compilation (2nd commit)
    """
    with newsource(fn2, issue_305_helper2, reexec=False):
        print(issue_305_helper2.test2(), issue_305_helper2.test2.py_func())
        assert issue_305_helper2.test2() == issue_305_helper2.test2.py_func()

test_fetch_latest_source()
test_no_auto_reload()


# -*- coding: utf-8 -*-

"""
Generate Python AST nodes and Cython pxd files.

Note: This file needs to be easy to externalize (factor out into another
project).
"""

from __future__ import print_function, division, absolute_import

import os
import ast
from textwrap import dedent
from functools import partial

from . import generator
from . import naming
from .classgen import Class, ClassCodegen
from .formatting import py_formatter, cy_formatter

root = os.path.dirname(os.path.abspath(__file__))
testdir = os.path.join(root, "tests")

#------------------------------------------------------------------------
# Code Formatting
#------------------------------------------------------------------------

py_preamble = """
import cython

# ASDL builtin types
# identifier = str
# string = basestring

class TypedProperty(object):
    '''Defines a class property that does a type check in the setter.'''

    def __new__(cls, ty, doc, default=None):
        rv = super(TypedProperty, cls).__new__(cls)
        cls.__init__(rv, ty, doc, default)
        return property(rv.getter, rv.setter, rv.deleter, doc)

    def __init__(self, ty, typename, attrname, default=None):
        self.propname = '_property_%d' % (id(self),)
        self.default = default
        if not isinstance(ty, tuple):
            ty = (ty,)
        self.ty = ty
        self.typename = typename
        self.attrname = attrname

    def getter(self, obj):
        return getattr(obj, self.propname, self.default)

    def setter(self, obj, new_val):
        if not isinstance(new_val, self.ty):
            ty = tuple(t.__name__ for t in self.ty)
            raise ValueError(
                "Invalid type for attribute '%s.%s', "
                "expected instance of type(s) %r "
                "(got %r)." % (self.typename, self.attrname,
                               ty, type(new_val).__name__))
        setattr(obj, self.propname, new_val)

    def deleter(self, obj):
        delattr(obj, self.propname)
"""

cy_preamble = """
cimport cython
from %s cimport GenericVisitor

# ctypedef str identifier
# ctypedef str string
# ctypedef bint bool
""" % (naming.interface,)

def format_field(classname, field):
    type = py_formatter.format_type(field.type)
    if field.opt:
        # types.NoneType is not in Python 3, type(None) should work
        # for both Python 2 and 3.
        type = "(%s, type(None))" % (type,)

    format_dict = dict(name=field.name, type=type, classname=classname)
    return ('%(classname)s.%(name)s = TypedProperty(%(type)s, '
                '"%(classname)s", "%(name)s")' % format_dict)

#------------------------------------------------------------------------
# Class Formatting
#------------------------------------------------------------------------

class PyClass(Class):
    """
    Generate Python AST classes.
    """

    def __str__(self):
        fieldnames = [str(field.name) for field in self.fields]
        fields = map(repr, fieldnames)
        properties = [format_field(self.name, field)
                          for field in self.fields] or ["pass"]

        if self.attributes is not None:
            attributes = map(repr, self.attributes)
            attributes = py_formatter.format_stats(",\n", 8, attributes)
            attributes = "_attributes = (" + attributes + ")"
        else:
            attributes = "# inherit _attributes"

        if fieldnames:
            initialize = ["self.%s = %s" % (name, name) for name in fieldnames]
        else:
            initialize = ["pass"]

        fmtstring = ", ".join("%s=%%s" % name for name in fieldnames)
        fmtargs = "(%s)" % ", ".join(py_formatter.get_fields(self.fields))

        format_dict = dict(
            name=self.name, base=self.base, doc=self.doc,
            fields      = py_formatter.format_stats(",\n", 8, fields),
            attributes  = attributes,
            properties  = py_formatter.format_stats("\n", 4, properties),
            params      = ", ".join(fieldnames),
            initialize  = py_formatter.format_stats("\n", 8, initialize),
            fmtstring   = fmtstring,
            fmtargs     = fmtargs,
        )

        return dedent('''
            class %(name)s(%(base)s):
                """
                %(doc)s
                """

                _fields = (
                    %(fields)s
                )

                %(attributes)s

                def __init__(self, %(params)s):
                    %(initialize)s

                def accept(self, visitor):
                    return visitor.visit_%(name)s(self)

                def __repr__(self):
                    return "%(name)s(%(fmtstring)s)" %% %(fmtargs)s

            if not cython.compiled:
                # Properties
                %(properties)s

        ''') % format_dict

class CyClass(Class):
    """
    Generate classes in pxd overlay.
    """

    def __str__(self):
        f = cy_formatter
        fields = ["cdef public %s %s" % (f.format_type(field.type), field.name)
                      for field in self.fields]
        fmtdict = {
            'name': self.name,
            'base': self.base,
            'fields': cy_formatter.format_stats("\n", 4, fields)
        }

        return dedent('''
            cdef class %(name)s(%(base)s):
                %(fields)s
                cpdef accept(self, GenericVisitor visitor)
        ''') % fmtdict

#------------------------------------------------------------------------
# Global Exports
#------------------------------------------------------------------------

def make_root_class(Class):
    return Class("AST", "object", doc="AST root node.", attributes=())

codegens = [
    ClassCodegen(naming.nodes + '.py', py_preamble,
                 PyClass, make_root_class(PyClass)),
    ClassCodegen(naming.nodes + '.pxd', cy_preamble,
                 CyClass, make_root_class(CyClass)),
]


if __name__ == '__main__':
    schema_filename = os.path.join(testdir, "testschema1.asdl")
    generator.generate_from_file(schema_filename, codegens, root)



# -*- coding: UTF-8 -*-

"""
numba --annotate
"""

from __future__ import print_function, division, absolute_import

import operator
from itertools import groupby
from collections import namedtuple

# ______________________________________________________________________

Program = namedtuple("Program", ["python_source", "intermediates"])
SourceIntermediate = namedtuple("SourceIntermediate", ["name", "linenomap",
                                                       "source"])
DotIntermediate = namedtuple("DotIntermediate", ["name", "dotcode"]) # graphviz
Source = namedtuple("Source", ["linemap", "annotations"])
Annotation = namedtuple("Annotation", ["type", "value"])

# ______________________________________________________________________

def build_linemap(func):
    import inspect
    import textwrap
    source = inspect.getsource(func)
    source = textwrap.dedent(source)
    lines = source.split('\n')
    if lines[-1] == '':
        lines = lines[0:-1]
    
    lineno = func.func_code.co_firstlineno
    linemap = {}
    for line in lines:
        linemap[lineno] = line
        lineno += 1

    return linemap

# ______________________________________________________________________

# Annotation types

A_type      = "Types"
A_c_api     = "Python C API"
A_numpy     = "NumPy"
A_errcheck  = "Error check"
A_objcoerce = "Coercion"
A_pycall    = "Python call"
A_pyattr    = "Python attribute"

# Annotation formatting

def format_annotations(annotations):
    adict = groupdict(annotations, 'type')
    for category, annotations in adict.items():
        vals = u" ".join([str(a.value) for a in annotations])
        yield u"%s: %s" % (category, vals)

groupdict = lambda xs, attr: dict(
    (k, list(v)) for k, v in groupby(xs, operator.attrgetter(attr)))

# ______________________________________________________________________


# -*- coding: UTF-8 -*-

"""
Capture IR emissions.
"""

from __future__ import print_function, division, absolute_import

import collections
from functools import partial
import llvm.core
from .annotate import SourceIntermediate, Source

# ______________________________________________________________________

class IRBuilder(object):
    def __init__(self, name, builder):
        self.name = name
        self.builder = builder
        self.captured = collections.defaultdict(list)
        self.pos = None

    def update_pos(self, pos):
        self.pos = pos

    def get_pos(self):
        return self.pos

    def __getattr__(self, attr):
        m = getattr(self.builder, attr)
        if not callable(m):
            return m

        def emit(*args, **kwargs):
            result = m(*args, **kwargs)
            self.captured[self.pos].append(result)
            return result

        return emit

# ______________________________________________________________________

def get_intermediate(ir_builder):
    "Get IR source from an IR builder as a SourceIntermediate"
    linenomap = collections.defaultdict(list)
    linemap = {}
    ir_lineno = 1

    filterer = filters.get(ir_builder.name, lambda x: x)
    ir_builder.captured = filterer(filter_unique(ir_builder.captured))

    for pos, instrs in sorted(ir_builder.captured.iteritems()):
        for instr in instrs:
            linenomap[pos].append(ir_lineno)
            linemap[ir_lineno] = str(instr)
            ir_lineno += 1

    source = Source(linemap, annotations=[])
    return SourceIntermediate(ir_builder.name, linenomap, source)

# ______________________________________________________________________

def filter_llvm(captured):
    for values in captured.values():
        fn = lambda llvm_value: isinstance(llvm_value, llvm.core.Instruction)
        blocks = collections.defaultdict(list)
        for llvm_value in filter(fn, values):
            blocks[llvm_value.basic_block].append(llvm_value)
        values[:] = order_llvm(blocks)
    return captured

def filter_unique(captured):
    for values in captured.values():
        def unique(item, seen=set()):
            found = item in seen
            seen.add(item)
            return not found

        values[:] = filter(unique, values)
    return captured

# ______________________________________________________________________

def order_llvm(blocks):
    """
    Put llvm instructions and basic blocks in the right order.

    :param blocks: { llvm_block : [llvm_instr] }
    :return: [llvm_line]
    """
    result = []
    if blocks:
        block, values = blocks.popitem()
        blocks[block] = values
        lfunc = block.function

        for block in lfunc.basic_blocks:
            if block in blocks:
                instrs = blocks[block]
                instrpos = dict(
                    (instr, i) for i, instr in enumerate(block.instructions))
                result.append(str(block.name) + ":")
                result.extend(sorted(instrs, key=instrpos.get))

    return result

# ______________________________________________________________________

filters = {
    "llvm": filter_llvm,
}


# -*- coding: utf-8 -*-

"""
Graph visualization for abitrary graphs. An attempt to unify all the many
different graphs we want to visualize (numba.control_flow.graphviz,
numba.minivect.graphviz, etc).
"""

from __future__ import print_function, division, absolute_import

import os
import ast
import logging
import textwrap
import subprocess
from itertools import chain

from numba.minivect.pydot import pydot

logger = logging.getLogger(__name__)

#------------------------------------------------------------------------
# Graphviz Mapper
#------------------------------------------------------------------------

class GraphvizGenerator(object):
    """
    Render an arbitrary graph as a graphviz tree.

    Nodes must be hashable.
    """

    counter = 0

    def __init__(self, graph_adaptor, graph_renderer,
                 graph_name, graph_type):
        self.graph_adaptor = graph_adaptor
        self.graph_renderer = graph_renderer

        self.graph = pydot.Dot(graph_name, graph_type=graph_type)

        # Graph Node -> PyDot Node
        self.seen = {}

        # { (source, dest) }
        self.seen_edges = set()

    # ______________________________________________________________________
    # Add to pydot graph

    def add_edge(self, source, dest):
        "Add an edge between two pydot nodes and set the colors"
        if (source, dest) in self.seen_edges:
            return

        self.seen_edges.add((source, dest))
        edge = pydot.Edge(self.seen[source], self.seen[dest])

        edge_label = self.graph_renderer.render_edge(source, dest)
        if edge_label is not None:
            edge.set_label(edge_label)

        self.graph.add_edge(edge)

    def create_node(self, node):
        "Create a graphviz node from the miniast node"
        label = self.graph_renderer.render(node)
        self.counter += 1
        pydot_node = pydot.Node(str(self.counter), label=label, shape='box')
        self.graph.add_node(pydot_node)
        return pydot_node

    # ______________________________________________________________________
    # Traverse Graph

    def dfs(self, node):
        "Visit children and add edges to their Graphviz nodes."
        if node in self.seen:
            return

        pydot_node = self.create_node(node)
        self.seen[node] = pydot_node

        for child in self.graph_adaptor.children(node):
            self.dfs(child)
            self.add_edge(node, child)

#------------------------------------------------------------------------
# Graph Adaptors
#------------------------------------------------------------------------

class GraphAdaptor(object):
    """
    Allow traversal of a foreign AST.
    """

    def children(self, node):
        "Return the children for this graph node"

#------------------------------------------------------------------------
# Graph Rendering
#------------------------------------------------------------------------

class GraphRenderer(object):
    """
    Allow traversal of a foreign AST.
    """

    def render(self, node):
        "Return the label for this graph node"

    def render_edge(self, source, dest):
        "Return the label for this edge or None"


#------------------------------------------------------------------------
# Create image from dot
#------------------------------------------------------------------------

def write_image(dot_output):

    prefix, ext = os.path.splitext(dot_output)
    png_output = prefix + '.png'

    fp = open(png_output, 'wb')
    try:
        p = subprocess.Popen(['dot', '-Tpng', dot_output],
                             stdout=fp.fileno(),
                             stderr=subprocess.PIPE)
        p.wait()
    except EnvironmentError as e:
        logger.warn("Unable to write png: %s (did you install the "
                    "'dot' program?). Wrote %s" % (e, dot_output))
    else:
        logger.warn("Wrote %s" % png_output)
    finally:
        fp.close()

#------------------------------------------------------------------------
# Entry points
#------------------------------------------------------------------------

def render(G, output_file, adaptor, renderer,
           graph_name="G", graph_type="digraph"):
    """
    G: The graph: [node]
    output_file: output dot file name
    adaptor: GraphAdaptor
    renderer: GraphRenderer
    """
    gen = GraphvizGenerator(adaptor, renderer, graph_name, graph_type)
    for root in G:
        gen.dfs(root)

    dotgraph = gen.graph
    # output_file, ext = os.path.splitext(output_file)
    # dotgraph.write(output_file + '.png', format='png')
    dotgraph.write(output_file)
    write_image(output_file)



# -*- coding: utf-8 -*-

"""
Postpasses over the LLVM IR.
The signature of each postpass is postpass(env, ee, lmod, lfunc) -> lfunc
"""

from __future__ import print_function, division, absolute_import

import llvmmath
from llvmmath import linking

default_postpasses = {}

def register_default(name):
    def dec(f):
        default_postpasses[name] = f
        return f
    return dec

# ______________________________________________________________________
# Postpasses

@register_default('math')
def postpass_link_math(env, ee, lmod, lfunc):
    "numba.math.* -> llvmmath.*"
    replacements = {}
    for lf in lmod.functions:
        if lf.name.startswith('numba.math.'):
            _, _, name = lf.name.rpartition('.')
            replacements[lf.name] = name
    del lf # this is dead after linking below

    default_math_lib = llvmmath.get_default_math_lib()
    linker = linking.get_linker(default_math_lib)
    linking.link_llvm_math_intrinsics(ee, lmod, default_math_lib,
                                      linker, replacements)
    return lfunc


# -*- coding: utf-8 -*-

"""
Numba annotators.
"""

from __future__ import print_function, division, absolute_import

import re

import numba
from numba import *
from numba import nodes

from numba.annotate import annotate
from numba.annotate.annotate import Annotation, A_c_api

logger = logging.getLogger(__name__)

# Taken from Cython/Compiler/Annotate.py
py_c_api = re.compile(u'(Py[A-Z][a-z]+_[A-Z][a-z][A-Za-z_]+)\(')
py_macro_api = re.compile(u'(Py[A-Z][a-z]+_[A-Z][A-Z_]+)\(')

# ______________________________________________________________________

class AnnotateTypes(object):

    def visit_Name(self, node):
        if isinstance(node, nodes.ExprNode):
            self.annotate(annotate.A_type, (node.id, str(node.type)))

# ______________________________________________________________________

def annotate_pyapi(llvm_intermediate, py_annotations):
    """
    Produce annotations for CPython C API calls from an LLVM SourceIntermediate
    """
    for py_lineno in llvm_intermediate.linenomap:
        count = 0
        for llvm_lineno in llvm_intermediate.linenomap[py_lineno]:
            line = llvm_intermediate.source.linemap[llvm_lineno]
            if re.search(py_c_api, line):
                count += 1

        if count:
            py_annotations[py_lineno].append(Annotation(A_c_api, count))



# -*- coding: utf-8 -*-

"""
Type lowering from a higher-level domain to a lower-level domain.
"""

from __future__ import print_function, division, absolute_import
from numba.typesystem import itypesystem

def find_matches(table, flags):
    "Find a lowering function from the flags of the type"
    matches = []
    for flag in flags:
        if flag in table:
            matches.append(flag)

    if len(matches) > 1:
        raise ValueError("Multiple matching flags: %s" % flags)
    elif matches:
        return matches[0]
    else:
        return None

def find_func(table, kind, flags, default=None):
    "Get a function form the table by resolving any indirections"
    if kind in table:
        flag = kind
    else:
        flag = find_matches(table, flags)
        if flag is None:
            return default

    while flag in table:
        if isinstance(table[flag], basestring):
            flag = table[flag]
        else:
            return table[flag]

    return default

def create_type_lowerer(table, domain, codomain):
    """
    Create a type lowerer from a domain to a codomain given a lowering table.
    """
    def convert_unit(domain, codomain, type):
        func = find_func(table, type.typename, type.flags)
        if func:
            return func(domain, codomain, type, ())
        else:
            return itypesystem.convert_unit(domain, codomain, type)

    def convert_para(domain, codomain, type, params):
        ctor = find_func(table, type.kind, type.flags, itypesystem.convert_para)
        # print("lowering...", type, ctor)
        return ctor(domain, codomain, type, params)

    return itypesystem.TypeConverter(domain, codomain, convert_unit, convert_para)

#------------------------------------------------------------------------
# Lowering functions
#------------------------------------------------------------------------

# ______________________________________________________________________
# unit types

def lower_object(domain, codomain, type, params):
    from numba import typedefs # hurr
    if type.is_array:
        return codomain.array_(*params)
    return codomain.pointer(typedefs.PyObject_HEAD)

def lower_string(domain, codomain, type, params):
    return codomain.pointer(codomain.char)

# ______________________________________________________________________
# parametrized types

def lower_function(domain, codomain, type, params):
    restype, args, name, is_vararg = params
    newargs = []

    for arg in args:
        if arg.is_struct or arg.is_function:
            arg = codomain.pointer(arg)
        newargs.append(arg)

    if restype.is_struct:
        newargs.append(codomain.pointer(restype))
        restype = codomain.void

    result = codomain.function(restype, newargs, name, is_vararg)
    # print("lowered", type, result, params)
    return result

def lower_extmethod(domain, codomain, type, params):
    return lower_function(domain, codomain, type, params[:4])

def lower_complex(domain, codomain, type, params):
    base_type, = params
    return codomain.struct_([('real', base_type), ('imag', base_type)])

def lower_to_pointer(domain, codomain, type, params):
    return codomain.pointer(params[0])

#------------------------------------------------------------------------
# Default Lowering Table
#------------------------------------------------------------------------

default_numba_lowering_table = {
    "object":           lower_object,
    # parametrized types
    "function":         lower_function,
    "complex":          lower_complex,
    # "array":            lower_array,
    "string":           lower_string,
    # "carray":           lower_to_pointer,
    "sized_pointer":    lower_to_pointer,
    "reference":        lower_to_pointer,
    "extmethod":        lower_extmethod,
    "known_pointer":    lower_to_pointer,
}

ctypes_lowering_table = {
    "object":           lambda dom, cod, type, params: cod.object_,
    "complex":          lower_complex,
    "array":            "object",
    # "string":           lambda dom, cod, type, params: ctypes.c_char_p,
    "sized_pointer":    lower_to_pointer,
    "reference":        lower_to_pointer,
    "extmethod":        lower_extmethod,
}


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import sys

from .testutils import generate_in_memory
from .. import generator, visitorgen, astgen, naming

try:
    import cython
    have_cython = True
except ImportError:
    have_cython = False

#------------------------------------------------------------------------
# Load modules
#------------------------------------------------------------------------

def load_testschema1():
    codegens = astgen.codegens + visitorgen.codegens
    file_allocator = generate_in_memory("testschema1.asdl", codegens)

    interface = generator.generate_module(file_allocator,
                                          naming.interface + '.py')
    sys.modules[naming.interface] = interface

    nodes = generator.generate_module(file_allocator, naming.nodes + '.py')
    sys.modules[naming.nodes] = nodes

    visitor = generator.generate_module(file_allocator,
                                        naming.visitor + '.py')
    transformer = generator.generate_module(file_allocator,
                                            naming.transformer + '.py')

    return nodes, visitor, transformer

if have_cython:
    nodes, visitor, transformer = load_testschema1()
else:
    class visitor(object):
        class Visitor(object):
            pass

    class transformer(object):
        class Transformer(object):
            pass

#------------------------------------------------------------------------
# Visitor testers
#------------------------------------------------------------------------

class TestVisitor(visitor.Visitor):

    def visit_Bar(self, node):
        print("Bar:", node.e1, node.e2)
        self.generic_visit(node)

    def visit_SomeExpr(self, node):
        print("SomeExpr:", node)

class TestTransformer(transformer.Transformer):

    def visit_SomeExpr(self, node):
        return nodes.SomeOtherExpr(node.n)

#------------------------------------------------------------------------
# Test funcs
#------------------------------------------------------------------------

def test_visitor():
    """
    >>> test_visitor()
    Bar: Bar(e1=SomeExpr(n=10), e2=SomeExpr(n=11)) SomeExpr(n=12)
    Bar: SomeExpr(n=10) SomeExpr(n=11)
    SomeExpr: SomeExpr(n=10)
    SomeExpr: SomeExpr(n=11)
    SomeExpr: SomeExpr(n=12)
    """
    e1 = nodes.SomeExpr(10)
    e2 = nodes.SomeExpr(11)
    expr = nodes.Bar(nodes.Bar(e1, e2), nodes.SomeExpr(12))
    TestVisitor().visit(expr)

def test_transformer():
    """
    >>> test_transformer()
    Bar(e1=Bar(e1=SomeOtherExpr(n=10), e2=SomeOtherExpr(n=11)), e2=SomeOtherExpr(n=12))
    """
    e1 = nodes.SomeExpr(10)
    e2 = nodes.SomeExpr(11)
    expr = nodes.Bar(nodes.Bar(e1, e2), nodes.SomeExpr(12))
    result = TestTransformer().visit(expr)
    print(result)


if __name__ == '__main__':
    import doctest
    if have_cython:
        sys.exit(0 if doctest.testmod().failed == 0 else 1)


# -*- coding: utf-8 -*-

"""
Some types to aid in type-based alias analysis. See numba/metadata.py.
"""

from __future__ import print_function, division, absolute_import

from numba.typesystem.types import NumbaType
from numba.typesystem import object_, npy_intp

class TBAAType(NumbaType):
    is_tbaa = True
    typename = "tbaa_type"
    argnames = ["name", "root"]

numpy_array = TBAAType("numpy array", object_)
numpy_shape = TBAAType("numpy shape", npy_intp.pointer())
numpy_strides = TBAAType("numpy strides", npy_intp.pointer())
numpy_ndim = TBAAType("numpy flags", npy_intp.pointer())
numpy_dtype = TBAAType("numpy dtype", object_)
numpy_base = TBAAType("numpy base", object_)
numpy_flags = TBAAType("numpy flags", npy_intp.pointer())



# -*- coding: utf-8 -*-

"""
>>> with_stat("foo", "bar")
Traceback (most recent call last):
      ...
NumbaError: ...: Only 'with python' and 'with nopython' is supported at this moment
"""

from __future__ import print_function, division, absolute_import

import numba
from numba.testing import testmod

@numba.autojit(warnstyle="simple")
def with_stat(fn, msg):
    with open(fn, 'w') as fp:
        fp.write(msg)

testmod()



# -*- coding: utf-8 -*-

"""
>>> f(1, 2)
(2, 1)
"""

from __future__ import print_function, division, absolute_import

import numba
from numba.testing import testmod

@numba.autojit
def f(a, b):
    for i in range(1):
        b, a = a, b

    return a, b

testmod()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from   numba import int_, jit
import numpy

@jit(argtypes=(int_[:],))
def test1(arr):
    u = 0
    for x in arr:
        u += len(arr)

    v = 0
    for y in arr:
        v += len(arr)

    return u + v

@jit(argtypes=(int_[:],))
def test2(arr):
    s = 0
    for i, x in enumerate(arr):
        s += i*x

    s2 = 0
    for i2, x2 in enumerate(arr, 1):
        s2 += i2*x2
    return s+s2

if __name__ == '__main__':
    arr = numpy.arange(1, 4)
    assert test1(arr) == test1.py_func(arr)
    assert test2(arr) == test2.py_func(arr)


"""Copyright (c) 2012, Daniele Mazzocchio  
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
* Neither the name of the developer nor the names of its contributors may be
  used to endorse or promote products derived from this software without
  specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."""

"""A light and fast template engine."""

import re


class Template(object):
    """ """

    def __init__(self, template):
        """Initialize class"""
        super(Template, self).__init__()
        self.template = template
        self.options = {"strip": False}

    def expand(self, namespace={}, **kw):
        """Return the expanded template string"""
        namespace.update(kw)
        output = []

        # Builtins
        namespace["echo"]   = lambda s: output.append(s)
        namespace["isdef"]  = lambda v: v in namespace
        namespace["setopt"] = lambda k, v: self.options.update({k: v})

        code = self._process(self._preprocess(self.template))
        eval(compile(code, "<string>", "exec"), namespace)
        return self._postprocess("".join(map(str, output)))

    def _preprocess(self, template):
        """Modify template string before code conversion"""
        # Replace inline ('%') blocks for easier parsing
        o = re.compile("(?m)^[ \t]*%((if|for|while|try).+:)")
        c = re.compile("(?m)^[ \t]*%(((else|elif|except|finally).*:)|(end\w+))")
        template = c.sub(r"<%:\g<1>%>", o.sub(r"<%\g<1>%>", template))

        # Replace (${x}) variables with '<%echo(x)%>'
        v = re.compile("\${([a-zA-Z0-9[\].\"\'_]+)}")
        template = v.sub(r"<%echo(\g<1>)%>\n", template)

        return template

    def _process(self, template):
        """Return the code generated from the template string"""
        code_blk = re.compile(r"<%(.*?)%>\n?", re.DOTALL)
        indent = 0
        code = []
        for n, blk in enumerate(code_blk.split(template)):
            # Replace '<\%' and '%\>' escapes
            blk = re.sub(r"<\\%", "<%", re.sub(r"%\\>", "%>", blk))
            # Unescape '%{}' characters
            blk = re.sub(r"\\(%|{|})", "\g<1>", blk)

            if not (n % 2):
                # Escape double-quote characters
                blk = re.sub(r"\"", "\\\"", blk)
                blk = (" " * (indent*4)) + 'echo("""{}""")'.format(blk)
            else:
                blk = blk.rstrip()
                if blk.lstrip().startswith(":"):
                    if not indent:
                        err = "unexpected block ending"
                        raise SyntaxError("Line {}: {}".format(n, err))
                    indent -= 1
                    if blk.startswith(":end"):
                        continue
                    blk = blk.lstrip()[1:]

                blk = re.sub("(?m)^", " " * (indent * 4), blk)
                if blk.endswith(":"):
                    indent += 1

            code.append(blk)

        if indent:
            err = "Reached EOF before closing block"
            raise EOFError("Line {}: {}".format(n, err))

        return "\n".join(code)

    def _postprocess(self, output):
        """Modify output string after variables and code evaluation"""
        if self.options["strip"]:
            output = re.sub("(?m)(^[ \t]+|[ \t]+$|(?<=[ \t])[ \t]+|^\n)", "",
                            output)

        return output



# -*- coding: utf-8 -*-

"""
>>> unpack_loop()
Traceback (most recent call last):
      ...
NumbaError: ...: Only a single target iteration variable is supported at the moment
"""

from __future__ import print_function, division, absolute_import

import numba
from numba.testing import testmod

@numba.autojit(warnstyle="simple")
def unpack_loop():
    x = [(1,2),(3,4)]
    for (a, b) in x:
        print(a + b)

testmod()



# -*- coding: utf-8 -*-

"""
Support for

    for i in numba.prange(...):
        ...

The implementation isn't particularly good, and should be greatly simplified
at some point.
"""

from __future__ import print_function, division, absolute_import

import ast
import copy
import warnings
import multiprocessing

try:
    NUM_THREADS = multiprocessing.cpu_count()
except NotImplementedError:
    warnings.warn("Unable to determine cpu count, assuming 2")
    NUM_THREADS = 2

import llvm.core
from llvm_cbuilder import *
from llvm_cbuilder import shortnames as C
from llvm_cbuilder import builder

import numba.decorators
from numba import *
from numba import error, visitors, nodes, templating
from numba.minivect import  minitypes
from numba import typesystem, pipeline
from numba.type_inference import infer
from numba.specialize.loops import unpack_range_args

from numba import threads

opmap = {
    # Unary
    ast.Invert: '~',
    ast.Not: None, # not supported
    ast.UAdd: '+',
    ast.USub: '-',

    # Binary
    ast.Add: '+',
    ast.Sub: '-',
    ast.Mult: '*',
    ast.Div: '/',
    ast.Mod: '%',
    ast.Pow: '**',
    ast.LShift: '<<',
    ast.RShift: '>>',
    ast.BitOr: '|',
    ast.BitXor: '^',
    ast.BitAnd: '&',
    ast.FloorDiv: '//',

    # Comparison
    ast.Eq: '==',
    ast.NotEq: '!=',
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.Gt: '>',
    ast.GtE: '>=',
    ast.Is: None,
    ast.IsNot: None,
    ast.In: None,
    ast.NotIn: None,
}


import logging
logger = logging.getLogger(__name__)

def get_reduction_op(op):
    # TODO: recognize invalid operators
    op = type(op)

    reduction_op = opmap[op]
    reduction_ops = {'-': '+', '/': '*'}
    if reduction_op in reduction_ops:
        reduction_op = reduction_ops[reduction_op]

    return reduction_op

def get_reduction_default(op):
    defaults = {
        '+': 0, '-': 0, '*': 1, '/': 1,
    }
    return defaults[op]

class VariableFindingVisitor(visitors.VariableFindingVisitor):
    "Find referenced and assigned ast.Name nodes"
    def __init__(self):
        super(VariableFindingVisitor, self).__init__()
        self.reductions = {}

    def register_assignment(self, node, target, op):
        if isinstance(target, ast.Name):
            self.assigned.add(target.id)

            if op is None:
                redop = op
            else:
                redop = get_reduction_op(op)

            if target.id in self.reductions:
                previous_op = self.reductions[target.id]
                if ((previous_op is None and op is not None) or
                        (previous_op is not None and op is None)):
                    raise error.NumbaError(
                            node, "Reduction variable %r may not be "
                                  "assigned to" % target.id)
                else:
                    if redop != previous_op:
                        raise error.NumbaError(
                            node, "Incompatible reduction operators: "
                                  "(%s and %s) for variable %r" % (
                                            op, previous_op, target.id))
            elif op:
                self.reductions[target.id] = redop

    def visit_Assign(self, node):
        if isinstance(node.targets[0], ast.Name):
            self.register_assignment(node, node.targets[0],
                                     getattr(node, 'inplace_op', None))


def create_prange_closure(env, prange_node, body, target):
    # Find referenced and assigned variables
    v = VariableFindingVisitor()
    v.visitlist(body)

    # Determine privates and reductions. Shared variables will be handled by
    # the closure support.
    privates = set(v.assigned) - set(v.reductions)
    reductions = v.reductions

    if isinstance(target, ast.Name) and target.id in reductions:
        # Remove target variable from reductions if present
        reductions.pop(target.id)
    privates.add(target.id)

    privates_struct_type = numba.struct([])
    privates_struct = ast.Name('__numba_privates', ast.Param())

    args = [privates_struct]
    func_def = ast.FunctionDef(name=templating.temp_name("prange_body"),
                               args=ast.arguments(args=args, vararg=None,
                                                  kwarg=None, defaults=[]),
                               body=copy.deepcopy(body),
                               decorator_list=[])

    # Update outlined prange body closure

    func_signature = void(privates_struct_type.ref())
    # func_signature.struct_by_reference = True
    need_closure_wrapper = False
    locals_dict = { '__numba_privates': privates_struct_type.ref() }

    func_env = env.translation.make_partial_env(
        func_def,
        func_signature=func_signature,
        need_closure_wrapper=need_closure_wrapper,
        locals=locals_dict,
    )

    # Update prange node
    prange_node.func_env = func_env
    prange_node.privates_struct_type = privates_struct_type
    prange_node.privates = privates
    prange_node.reductions = reductions
    prange_node.func_def = func_def


prange_template = """
{{func_def}}
%s # function name; avoid warning about unused variable

$pack_struct

$nsteps = ({{stop}} - {{start}}) / ({{step}} * {{num_threads}})
for $i in range({{num_threads}}):
    $temp_struct.__numba_closure_scope = {{closure_scope}}
    $temp_struct.__numba_start = {{start}} + $i * {{step}} * $nsteps
    $temp_struct.__numba_stop = $temp_struct.__numba_start + {{step}} * $nsteps
    $temp_struct.__numba_step = {{step}}
    $contexts[$i] = $temp_struct

    # print "temp struct", $temp_struct.__numba_start, \
    #       $temp_struct.__numba_stop, {{step}}, $nsteps

# Undo any truncation, don't use $i here, range() doesn't
# have py semantics yet
$contexts[{{num_threads}} - 1].__numba_stop = {{stop}}

# print "invoking..."
{{invoke_and_join_threads}}

$lastprivates = $contexts[{{num_threads}} - 1]
"""

def kill_attribute_assignments(env, prange_node, temporaries):
    """
    Remove attribute assignments from the list of statements that need to
    be resolved before type inference.
    """
    func_env = env.translation.crnt
    kill_set = func_env.kill_attribute_assignments
    kill_set.update(temporaries)
    kill_set.update(prange_node.privates)
    kill_set.update(prange_node.reductions)

def rewrite_prange(env, prange_node, target, locals_dict, closures_dict):
    func_def = prange_node.func_def
    struct_type = prange_node.privates_struct_type

    templ = templating.TemplateContext(env.context,
                                       prange_template % func_def.name)

    # Allocate context for each thread
    num_threads = NUM_THREADS
    contexts_array_type = minitypes.CArrayType(struct_type,
                                               num_threads)

    # Create variables for template substitution
    nsteps = templ.temp_var('nsteps')
    temp_i = templ.temp_var('i', int32)
    contexts = templ.temp_var('contexts', contexts_array_type)
    temp_struct = templ.temp_var('temp_struct', struct_type)
    lastprivates = templ.temp_var("lastprivates")

    pack_struct = templ.code_var('pack_struct')

    if isinstance(target, ast.Name):
        target_name = target.id
        struct_type.add_field(target_name, Py_ssize_t)
    else:
        raise error.NumbaError(
            prange_node, "Only name target for prange is currently supported")

    # Create code for reductions and (last)privates
    for name, reduction_op in prange_node.reductions.iteritems():
        default = get_reduction_default(reduction_op)
        pack_struct.codes.append("%s.%s = %s" % (temp_struct, name,
                                                 default))

    # Update struct type with closure scope, index variable, start,
    # stop and step
    struct_type.add_field('__numba_closure_scope', void.pointer())
    struct_type.add_field('__numba_start', npy_intp)
    struct_type.add_field('__numba_stop', npy_intp)
    struct_type.add_field('__numba_step', npy_intp)

    # Interpolate code and variables and run type inference
    # TODO: UNDO monkeypatching
    func_def.type = prange_node.func_env.func_signature
    func_def.is_prange_body = True
    func_def.prange_node = prange_node

    num_threads_node = nodes.const(num_threads, Py_ssize_t)

    invoke = InvokeAndJoinThreads(env, contexts=contexts.node,
                                  func_def_name=func_def.name,
                                  struct_type=struct_type,
                                  target_name=target_name,
                                  num_threads=num_threads_node,
                                  closures=closures_dict)

    closure_scope = nodes.ClosureScopeLoadNode()
    subs = dict(
        func_def=func_def,
        func_def_name=func_def.name,
        closure_scope=closure_scope,
        invoke_and_join_threads=invoke,
        num_threads=num_threads_node,
        start=nodes.UntypedCoercion(prange_node.start, Py_ssize_t),
        stop=nodes.UntypedCoercion(prange_node.stop, Py_ssize_t),
        step=nodes.UntypedCoercion(prange_node.step, Py_ssize_t),
    )

    tree = templ.template(subs)

    temporaries = {}
    templ.update_locals(temporaries)
    locals_dict.update(temporaries)

    kill_attribute_assignments(env, prange_node, temporaries)

    # TODO: Make this an SSA variable
    locals_dict[target_name] = Py_ssize_t

    prange_node.target = target
    prange_node.num_threads_node = num_threads_node #.clone
    prange_node.template_vars = {
                                  'contexts': contexts,
                                  'i': temp_i,
                                  'lastprivates': lastprivates,
                                  'nsteps': nsteps,
                                }

    # print(templ.substituted_template)
    return tree

def typeof(name, expr):
    return "__numba_typeof(%s, %s)" % (name, expr)

def assign(name, expr):
    return "%s = %s" % (name, typeof(name, expr))

post_prange_template = """
#for $i in range({{num_threads}}):
#    {{invoke_thread}}

#for $i in range({{num_threads}}):
#    {{join_thread}}

# print "performing reductions"
for $i in range({{num_threads}}):
    $reductions

# print "unpacking lastprivates"
$unpack_struct
"""

def perform_reductions(context, prange_node):
    templ = templating.TemplateContext(context, post_prange_template)

    unpack_struct, reductions = templ.code_vars('unpack_struct', 'reductions')
    reductions.sep = "; "

    getvar = prange_node.template_vars.get
    templ.add_variable(getvar("i"))

    # Create code for reductions and (last)privates
    for name, reduction_op in prange_node.reductions.iteritems():
        # Generate: x += contexts[i].x
        expr = "%s %s %s[%s].%s" % (name, reduction_op,
                                    getvar("contexts"), getvar("i"),
                                    name)
        reductions.codes.append(assign(name, expr))

    target_name = ""
    if isinstance(prange_node.target, ast.Name):
        target_name = prange_node.target.id

    for name in prange_node.privates:
        # Generate: x += contexts[num_threads - 1].x
        expr = "%s.%s" % (getvar("lastprivates"), name)
        assmnt = assign(name, expr)

        if name == target_name:
            assmnt = "if %s > 0: %s" % (getvar("nsteps"), assmnt)

        unpack_struct.codes.append(assmnt)

    substitutions = { "num_threads": prange_node.num_threads_node }
    result = templ.template(substitutions)

    # print(templ.substituted_template)
    return result

#------------------------------------------------------------------------
# prange nodes and types
#------------------------------------------------------------------------

class PrangeType(typesystem.NumbaType):
    is_prange = True
    is_range = True

class PrangeNode(nodes.ExprNode):
    """
    Prange node. This replaces the For loop iterator in the initial stage.
    After type inference and before closure type inference it replaces the
    entire loop.
    """

    _fields = ['start', 'stop', 'step']

    func_def = None                 # outlined prange closure body
    privates_struct_type = None     # numba.struct(var_name=var_type)
    privates = None                 # set([var_name])
    reductions = None               # { var_name: reduction_op }
    target = None                   # Target iteration variable

    num_threads_node = None         # num_threads CloneNode
    template_vars = None            # { "template_var_name": $template_var }

    def __init__(self, start, stop, step, **kwargs):
        super(PrangeNode, self).__init__(**kwargs)
        self.start = start
        self.stop = stop
        self.step = step
        self.type = PrangeType()


class InvokeAndJoinThreads(nodes.UserNode):
    """
    contexts
        contexts array node (array of privates structs)
    num_threads
        num threads node

    func_def_name
        name of outlined prange body function
    target_name
        name of iteration target variable (e.g. 'i')
    struct_type
        privates struct type
    closures
        { closure_name : closure_node }
    """

    _fields = ['contexts', 'num_threads']

    def __init__(self, env, **kwargs):
        super(InvokeAndJoinThreads, self).__init__(**kwargs)
        self.env = env

    def infer_types(self, type_inferer):
        type_inferer.visitchildren(self)
        return self

    def build_wrapper(self, codegen):
        closure_type = self.closures[self.func_def_name].type
        lfunc = closure_type.closure.lfunc
        lfunc_pointer = closure_type.closure.lfunc_pointer

        KernelWrapper, RunThreadPool = get_threadpool_funcs(
            codegen.context,
            self.struct_type,
            self.target_name,
            lfunc,
            lfunc_pointer,
            closure_type.signature,
            self.num_threads,
            codegen.llvm_module)

        kernel_wrapper = nodes.LLVMCBuilderNode(self.env, KernelWrapper, None)
        run_threadpool = nodes.LLVMCBuilderNode(self.env, RunThreadPool, None,
                                                dependencies=[kernel_wrapper])
        lfunc_run = codegen.visit(run_threadpool)
        return lfunc_run

    def codegen(self, codegen):
        contexts = codegen.visit(self.contexts)
        num_threads = codegen.visit(self.num_threads.coerce(int_))

        lfunc_run = self.build_wrapper(codegen)
        codegen.builder.call(lfunc_run, [contexts, num_threads])
        return None


class TypeofNode(nodes.UserNode):

    _fields = ["name", "expr"]

    def __init__(self, name, expr):
        self.name = name
        self.expr = expr

    def infer_types(self, type_inferer):
        if type_inferer.analyse:
            return type_inferer.visit(self.expr)

        self.name = type_inferer.visit(self.name)
        self.type = self.name.variable.type
        return self


def make_privates_struct_type(privates_struct_type, names):
    """
    Update the struct of privates and reductions once we know the
    field types.
    """
    fielddict = dict((name.id, name.variable.type) for name in names)
    fielddict.update(privates_struct_type.fielddict)
    fields = numba.struct(**fielddict).fields

    privates_struct_type.fields[:] = fields
    # privates_struct_type.fielddict = fielddict
    privates_struct_type.update_mutated()


class VariableTypeInferingNode(nodes.UserNode):

    _fields = ["names", "pre_prange_code"]

    def __init__(self, variable_names, privates_struct_type):
        super(VariableTypeInferingNode, self).__init__()
        self.privates_struct_type = privates_struct_type

        self.names = []
        for varname in variable_names:
            self.names.append(ast.Name(id=varname, ctx=ast.Load()))

    def infer_types(self, type_inferer):
        type_inferer.visitchildren(self)
        make_privates_struct_type(self.privates_struct_type, self.names)
        return None


#------------------------------------------------------------------------
# prange visitors
#------------------------------------------------------------------------

class PrangeExpander(visitors.NumbaTransformer):
    """
    Rewrite 'for i in prange(...): ...' before the control flow pass.
    """

    prange = 0

    def visit_FunctionDef(self, node):
        if self.func_level == 0:
            node = self.visit_func_children(node)

        return node

    def match_global(self, node, expected_value):
        if isinstance(node, ast.Name) and node.id not in self.local_names:
            value = self.func_globals.get(node.id, None)
            return value is expected_value
        return False

    def is_numba_prange(self, node):
        return (self.match_global(node, prange) or
                (isinstance(node, ast.Attribute) and node.attr == "prange" and
                 self.match_global(node.value, numba)))

    def visit_Call(self, node):
        if self.is_numba_prange(node.func):
            infer.no_keywords(node)
            start, stop, step = self.visitlist(unpack_range_args(node))
            node = PrangeNode(start, stop, step)

        self.visitchildren(node)
        return node

    def error_check_prange(self, node):
        if self.prange:
            raise error.NumbaError(node, "Cannot nest prange")
        if node.orelse:
            raise error.NumbaError(node.orelse,
                                   "Else clause to prange not yet supported")

    def visit_For(self, node):
        node.iter = self.visit(node.iter)

        if not isinstance(node.iter, PrangeNode):
            self.visitchildren(node)
            return node

        self.error_check_prange(node)
        node.target = self.visit(node.target)

        self.prange += 1
        node.body = self.visitlist(node.body)
        self.prange -= 1

        # Create prange closure
        prange_node = node.iter
        create_prange_closure(self.env, prange_node, node.body, node.target)

        # setup glue code
        pre_loop = rewrite_prange(self.env, prange_node, node.target,
                                  self.locals, self.closures)
        post_loop = perform_reductions(self.context, prange_node)

        # infer glue code at the right place
        pre_loop_dont_infer = nodes.dont_infer(pre_loop)
        pre_loop_infer_now = nodes.infer_now(pre_loop, pre_loop_dont_infer)

        # infer the type of the struct of privates right after the loop
        allprivates = set(prange_node.privates) | set(prange_node.reductions)
        type = prange_node.privates_struct_type
        infer_privates_struct = VariableTypeInferingNode(allprivates, type)

        # Signal that we now have additional local variables
        self.invalidate_locals()

        return ast.Suite(body=[
            pre_loop_dont_infer,
            node,
            infer_privates_struct,
            pre_loop_infer_now,
            post_loop])


class PrangeCleanup(visitors.NumbaTransformer):
    """
    Clean up outlined prange loops after type inference (removes them entirely).
    """

    def visit_For(self, node):
        if not node.iter.variable.type.is_prange:
            self.visitchildren(node)
            return node

        nodes.delete_control_blocks(node, self.ast.flow)
        return None


class PrangePrivatesReplacer(visitors.NumbaTransformer):
    """
    Rewrite private variables to accesses on the privates before
    closure type inference (closure type inference of the outer function, and
    type inference (and control flow analysis) of the inner function).
    """

    in_prange_closure = 0

    def visit_FunctionDef(self, node):
        """
        Analyse immedidate prange functions (not ones in closures).

        Don't re-analyze prange functions when the prange function closures
        themselves are compiled.
        """
        if getattr(node, 'is_prange_body', False) and self.func_level == 0:
            prange_node = node.prange_node
            self.privates_struct_type = prange_node.privates_struct_type

            node.body = [nodes.WithNoPythonNode(body=node.body)]

            self.in_prange_closure += 1
            self.visit_func_children(node)
            self.in_prange_closure -= 1

            self.invalidate_locals(node)
            self.invalidate_locals()
        else:
            self.visit_func_children(node)

        return node

    def visit_Name(self, node):
        if self.in_prange_closure:
            if node.id in self.privates_struct_type.fielddict:
                privates_struct = ast.Name('__numba_privates', ast.Load()) #node.ctx)
                result = ast.Attribute(value=privates_struct,
                                       attr=node.id,
                                       ctx=node.ctx)
                return result

        return node

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id == "__numba_typeof":
            return TypeofNode(node.args[0], node.args[1])

        self.visitchildren(node)
        return node


#----------------------------------------------------------------------------
# LLVM cbuilder prange utilities
#----------------------------------------------------------------------------

_count = 0
def get_threadpool_funcs(context, context_struct_type, target_name,
                         lfunc, lfunc_pointer, signature,
                         num_threads, llvm_module):
    """
    Get functions to run the closure in separate threads.

        context:
            the Numba/Minivect context

        context_struct_type:
            the struct type holding all private and reduction variables
    """
    global _count
    _count += 1

    context_cbuilder_type = builder.CStruct.from_numba_struct(
                                    context, context_struct_type)
    context_p_ltype = context_struct_type.pointer().to_llvm(context)

    class KernelWrapper(CDefinition):
        """
        Implements a prange kernel wrapper that is invoked in each thread.
        Implements:

            for i in range(start, stop, step):
                worker(closure_scope, privates)
        """

        _name_ = "prange_kernel_wrapper_%d" % _count
        _argtys_ = [
            ('context', C.void_p),
        ]

        def __init__(self, dependencies, ldependencies, **kwargs):
            super(KernelWrapper, self).__init__(**kwargs)

        def dispatch(self, context_struct_p, context_getfield,
                     lfunc, lfunc_pointer):
            """
            Call the closure with the closure scope and context arguments.
            We don't directly call the lfunc since there are linkage issues.
            """
            if signature.args[0].is_closure_scope:
                llvm_object_type = object_.to_llvm(context)
                closure_scope = context_getfield('closure_scope')
                closure_scope = closure_scope.cast(llvm_object_type)
                args = [closure_scope, context_struct_p]
            else:
                args = [context_struct_p]

            # Get the LLVM arguments
            llargs = [arg.handle for arg in args]

            # Get the LLVM pointer to the function
            lfunc_pointer = llvm.core.Constant.int(Py_uintptr_t.to_llvm(context),
                                                   lfunc_pointer)
            lfunc_pointer = self.builder.inttoptr(lfunc_pointer, lfunc.type)

            self.builder.call(lfunc_pointer, llargs)

        def body(self, context_p):
            context_struct_p = context_p.cast(context_p_ltype)
            context_struct = context_struct_p.as_struct(context_cbuilder_type)

            def context_getfield(name):
                "Get a field named __numba_<name>"
                return getattr(context_struct, '__numba_' + name)

            start = self.var(C.npy_intp, context_getfield('start'))
            stop = self.var(C.npy_intp, context_getfield('stop'))
            step = self.var(C.npy_intp, context_getfield('step'))

            length = stop - start
            nsteps = self.var(C.npy_intp, length / step)

            zero = self.constant(C.npy_intp, 0)

            with self.ifelse(length % step != zero) as ifelse:
                with ifelse.then():
                    nsteps += self.constant(C.npy_intp, 1)

            # self.debug("start", start, "stop", stop, "step", step)
            with self.for_range(nsteps) as (loop, i):
                getattr(context_struct, target_name).assign(start)
                self.dispatch(context_struct_p, context_getfield,
                              lfunc, lfunc_pointer)
                start += step

            self.ret()

        def specialize(self, *args, **kwargs):
            self._name_ = "__numba_kernel_wrapper_%s" % lfunc.name

    class RunThreadPool(CDefinition, threads.ParallelMixin):
        """
        Function that spawns the thread pool.
        """

        _name_ = "invoke_prange_%d" % _count
        _argtys_ = [
            ('contexts', context_p_ltype),
            ('num_threads', C.int),
        ]

        def __init__(self, dependencies, ldependencies, **kwargs):
            self.kernel_wrapper, = dependencies
            super(RunThreadPool, self).__init__(**kwargs)

        def body(self, contexts, num_threads):
            callback = self.kernel_wrapper.pointer
            callback = self.constant(Py_uintptr_t.to_llvm(context), callback)
            callback = callback.cast(C.void_p)
            self._dispatch_worker(callback, contexts, num_threads)
            self.ret()

        def specialize(self, *args, **kwargs):
            self._name_ = "__numba_run_threadpool_%s" % lfunc.name

    return KernelWrapper, RunThreadPool

#----------------------------------------------------------------------------
# The actual prange function
#----------------------------------------------------------------------------

def prange(start=0, stop=None, step=1):
    if stop is None:
        stop = start
        start = 0
    return range(start, stop, step)

# numba.prange = prange



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba import autojit

@autojit(nopython=True)
def test_chr(x):
    return chr(x)

assert test_chr(97) == b'a'


# -*- coding: UTF-8 -*-
from __future__ import print_function, division, absolute_import
import sys
from functools import partial
from itertools import chain
from collections import namedtuple

from numba.lexing import lex_source
from .annotate import format_annotations

WIDTH = 40
ANNOT_SEP = "-"

Emitter = namedtuple("Emitter", ["emit", "emitline"])
lex = partial(lex_source, output="console")

# ______________________________________________________________________

def render(annotation_blocks, emit=sys.stdout.write,
           intermediate_names=(), inline=True):
    """
    Render a Program as text.

    :param intermediate_names: [intermediate_name], e.g. ["llvm"]
    :param inline: whether to display intermediate code inline
    """
    indent = 8
    emitline = lambda indent, s: emit(u" " * indent + s + u"\n")
    emitter = Emitter(emit, emitline)

    for i, block in enumerate(annotation_blocks):
        python_source = block['python_source']
        intermediates = block['intermediates']

        if intermediates:
            irs = [i for i in intermediates if i.name in intermediate_names]
        else:
            irs = None

        # Render main source
        render_source(python_source, emitter, indent,
                      irs if inline and irs else [])

        if not inline and irs:
            # Render IRs seperately
            for irname, linenomap, ir_source in irs:
                emitter.emitline(0, irname.center(80, "="))
                render_source(ir_source, emitter, indent, [], linenomap)
            emitter.emitline(0, "=" * 80)

def render_source(source, emitter, indent, intermediates, linenomap=None):
    """
    Print a Source and its annotations. Print any given Intermediates inline.
    """
    if linenomap:
        indent += 8
        headers = {}
        for py_lineno, ir_linenos in linenomap.items():
            for ir_lineno in ir_linenos:
                headers[ir_lineno] = u"%4d |  " % py_lineno

        header = lambda lineno: headers.get(lineno, u"     |  ")
    else:
        header = lambda lineno: u""

    _render_source(source, emitter, indent, intermediates, header)


def _render_source(source, emitter, indent, intermediates, header=None):
    for lineno in sorted(source.linemap.iterkeys()):
        if header:
            emitter.emit(header(lineno))
        line = lex(source.linemap[lineno])
        emitter.emitline(0, u"%4d    %s" % (lineno, line))

        annots = format_annotations(source.annotations.get(lineno, []))
        irs = _gather_text_intermediates(intermediates, lineno)
        lines = list(chain(annots, irs))
        if not lines:
            continue

        # Print out annotations
        linestart = indent + len(source.linemap[lineno]) - len(source.linemap[lineno].lstrip())
        emitter.emitline(linestart + 2, u"||".center(WIDTH, ANNOT_SEP))
        for line in lines:
            emitter.emitline(linestart + 2, line)
        emitter.emitline(linestart + 2, u"||".center(WIDTH, ANNOT_SEP))

def _gather_text_intermediates(intermediates, lineno):
    for irname, linenomap, ir_source in intermediates:
        ir_linenos = linenomap.get(lineno, [])
        if not ir_linenos:
            continue
        yield irname.center(WIDTH, "_")
        for ir_lineno in ir_linenos:
            yield lex(ir_source.linemap[ir_lineno], irname)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import ctypes
import unittest

import numba as nb
from numba import jit, list_, tuple_, object_, int_, sized_pointer, npy_intp

import numpy as np

A = np.empty((5, 6))
shape_t = sized_pointer(npy_intp, 2)

# ______________________________________________________________________

def unpack(x):
    x, y = x
    return x * y

# ______________________________________________________________________

class Iterable(object):
    """I don't work yet"""

    def __iter__(self):
        return iter((5, 6))

class Sequence(object):
    """I work"""

    def __getitem__(self, idx):
        return [5, 6][idx]

# ______________________________________________________________________

class TestUnpacking(unittest.TestCase):

    def test_unpacking(self):
        lunpack  = jit(int_(list_(int_, 2)))(unpack)
        tunpack  = jit(int_(tuple_(int_, 2)))(unpack)
        tounpack = jit(int_(tuple_(object_, 2)))(unpack)
        iunpack  = jit(int_(object_))(unpack)
        sunpack  = jit(int_(object_))(unpack)
        punpack  = jit(int_(shape_t), wrap=False)(unpack)

        self.assertEqual(lunpack([5, 6]), 30)
        self.assertEqual(tunpack((5, 6)), 30)
        self.assertEqual(tounpack((5, 6)), 30)
        # self.assertEqual(iunpack(Iterable()), 30)
        self.assertEqual(sunpack(Sequence()), 30)

        c_punpack = nb.addressof(punpack)
        self.assertEqual(c_punpack(A.ctypes.shape), 30)

if __name__ == "__main__":
    unittest.main()



import ctypes
import unittest
from collections import namedtuple
from functools import partial
import numba
from numba import *
from numba import ndarray_helpers
import llvm.core as lc

# ______________________________________________________________________

ArrayType = numba.struct([('data', double.pointer()),
                          ('shape', int64.pointer()),
                          ('strides', int64.pointer())])

Int32 = lc.Type.int(32)
const = partial(lc.Constant.int, Int32)
zero = const(0)
one  = const(1)
two  = const(2)

def ptr_at(builder, ptr, idx):
    return builder.gep(ptr, [const(idx)])

def load_at(builder, ptr, idx):
    return builder.load(ptr_at(builder, ptr, idx))

def store_at(builder, ptr, idx, val):
    builder.store(val, ptr_at(builder, ptr, idx))

class MyArray(object):
    """
    Internal array class for a double(:, 10) array.
    """

    def __init__(self, array_ptr, builder):
        self.array_ptr = array_ptr
        self.builder = builder
        self.nd = 2

    @classmethod
    def from_type(cls, llvm_dtype):
        return ArrayType.pointer().to_llvm()

    @property
    def data(self):
        dptr = self.builder.gep(self.array_ptr, [zero, zero])
        return self.builder.load(dptr)

    @property
    def shape_ptr(self):
        result = self.builder.gep(self.array_ptr, [zero, one])
        result = self.builder.load(result)
        return result

    @property
    def strides_ptr(self):
        result = self.builder.gep(self.array_ptr, [zero, two])
        return self.builder.load(result)

    @property
    def shape(self):
        return self.preload(self.shape_ptr, self.nd)

    @property
    def strides(self):
        return self.preload(self.strides_ptr, self.nd)

    @property
    def ndim(self):
        return const(self.nd)

    @property
    def itemsize(self):
        raise NotImplementedError

    def preload(self, ptr, count=None):
        assert count is not None
        return [load_at(self.builder, ptr, i) for i in range(count)]

    def getptr(self, *indices):
        const = partial(lc.Constant.int, indices[0].type)
        offset = self.builder.add(
            self.builder.mul(indices[0], const(10)), indices[1])
        data_ptr_ty = lc.Type.pointer(lc.Type.double())
        dptr_plus_offset = self.builder.gep(self.data, [offset])
        return self.builder.bitcast(dptr_plus_offset, data_ptr_ty)

ndarray_helpers.Array.register(MyArray)

# ______________________________________________________________________
# Test functions

CtypesArray = ArrayType.to_ctypes()

@jit(void(double[:, :]), array=MyArray, wrap=False, nopython=True)
def use_array(A):
    """simple test function"""
    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            A[i, j] = i * A.shape[1] + j

@jit(object_(double[:, :]), array=MyArray, wrap=False)
def get_attributes(A):
    return A.shape[0], A.shape[1], A.strides[0], A.strides[1]

# ______________________________________________________________________

# Ctypes functions
c_use_array      = numba.addressof(use_array)
c_get_attributes = numba.addressof(get_attributes)

c_use_array.argtypes      = [ctypes.POINTER(CtypesArray)]
c_get_attributes.argtypes = [ctypes.POINTER(CtypesArray)]

# ______________________________________________________________________
# Utils

Array = namedtuple('Array', ['handle', 'array', 'data', 'shape', 'strides'])

def make_array():
    """Make a double[*, 10] ctypes-allocated array"""
    empty = lambda c_type, args: ctypes.cast(
        (c_type * len(args))(*args), ctypes.POINTER(c_type))

    data    = empty(ctypes.c_double, [0] * 50)
    shape   = empty(ctypes.c_int64,  [5, 10])
    strides = empty(ctypes.c_int64,  [10 * 8, 8]) # doubles!
    array   = CtypesArray(data, shape, strides)

    return Array(ctypes.pointer(array), array, data, shape, strides)

# ______________________________________________________________________
# Tests...

class TestArray(unittest.TestCase):

    def test_indexing(self):
        arr = make_array()
        c_use_array(arr.handle)
        for i in range(50):
            assert arr.data[i] == float(i), (arr.data[i], i)

    def test_attributes(self):
        arr = make_array()
        result = c_get_attributes(arr.handle)
        assert result == (5, 10, 80, 8), result


if __name__ == "__main__":
    # TestArray('test_attributes').debug()
    # TestArray('test_indexing').debug()
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba import *

@autojit
def zip1(L1, L2):
    """
    >>> zip1(range(2), range(5, 8))
    [(0, 5), (1, 6)]
    """
    return list(zip(L1, L2))

@autojit
def zip2(L1, L2, L3):
    """
    >>> zip2(range(2), range(5, 8), range(9, 13))
    [(0, 5, 9), (1, 6, 10)]
    """
    return list(zip(L1, L2, L3))

@autojit
def ziploop1(L1, L2):
    """
    >>> ziploop1(range(2), range(5, 8))
    0 5
    1 6
    """
    for i, j in zip(L1, L2):
        print(i, j)

@autojit
def ziploop2(L1, L2, L3):
    """
    >>> ziploop2(range(2), range(5, 8), range(9, 13))
    0 5 9
    1 6 10
    """
    for i, j, k in zip(L1, L2, L3):
        print(i, j, k)


if __name__ == '__main__':
    import numba
    numba.testing.testmod()



# -*- coding: utf-8 -*-
"""
Type functions for Python builtins.
"""
from __future__ import print_function, division, absolute_import

import warnings

import ast
from numba import *
from numba import nodes
from numba import error
# from numba import function_util
from numba.symtab import Variable
from numba import typesystem
from numba.typesystem import get_type

from numba.type_inference.modules import utils

#----------------------------------------------------------------------------
# Utilities
#----------------------------------------------------------------------------

register_builtin = utils.register_with_argchecking

def cast(node, dst_type):
    if len(node.args) == 0:
        return nodes.ConstNode(0, dst_type)
    else:
        return nodes.CoercionNode(node.args[0], dst_type=dst_type)

#----------------------------------------------------------------------------
# Type Functions for Builtins
#----------------------------------------------------------------------------

# TODO: add specializer functions to insert coercions before late specialization
# TODO: don't rewrite AST here

@register_builtin((1, 2, 3), can_handle_deferred_types=True)
def range_(typesystem, node, start, stop, step):
    node.variable = Variable(typesystem.range_)
    node.args = nodes.CoercionNode.coerce(node.args, dst_type=Py_ssize_t)
    return node

if not PY3:
    @register_builtin((1, 2, 3), can_handle_deferred_types=True)
    def xrange_(typesystem, node, start, stop, step):
        return range_(typesystem, node, start, stop, step)

@register_builtin(1)
def len_(typesystem, node, obj):
    # Simplify len(array) to ndarray.shape[0]
    argtype = get_type(obj)
    if argtype.is_array:
        shape_attr = nodes.ArrayAttributeNode('shape', node.args[0])
        new_node = nodes.index(shape_attr, 0)
        return new_node
    elif argtype.is_string:
        return nodes.CoercionNode(nodes.typednode(node, size_t), Py_ssize_t)

    return Py_ssize_t # Object call

@register_builtin((0, 1, 2), can_handle_deferred_types=True)
def _int(typesystem, node, x, base, dst_type=int_):
    # Resolve int(x) and float(x) to an equivalent cast

    if len(node.args) < 2:
        return cast(node, dst_type)

    node.variable = Variable(dst_type)
    return node

if not PY3:
    @register_builtin((0, 1, 2), can_handle_deferred_types=True)
    def _long(typesystem, node, x, base):
        return _int(typesystem, node, x, base)

@register_builtin((0, 1), can_handle_deferred_types=True)
def _float(typesystem, node, x):
    return cast(node, double)

@register_builtin((0, 1), can_handle_deferred_types=True)
def _bool(context, node, x):
    return cast(node, bool_)

@register_builtin((0, 1, 2), can_handle_deferred_types=True)
def complex_(typesystem, node, a, b):
    if len(node.args) == 2:
        args = nodes.CoercionNode.coerce(node.args, double)
        return nodes.ComplexNode(real=args[0], imag=args[1])
    else:
        return cast(node, complex128)

def abstype(argtype):
    if argtype.is_complex:
        result_type = argtype.base_type
    elif argtype.is_float or argtype.is_int:
        result_type = argtype
    else:
        result_type = object_

    return result_type

@register_builtin(1)
def abs_(typesystem, node, x):
    argtype = typesystem.promote(long_, get_type(x))
    dst_type = abstype(argtype)
    node.variable = Variable(dst_type)
    node.args = [nodes.CoercionNode(x, argtype)]
    return node

@register_builtin((2, 3))

def pow_(typesystem, node, base, exponent, mod=None):
    if mod:
        warnings.warn(
            "pow() with modulo (third) argument not natively supported")
        return nodes.call_pyfunc(pow, [base, exponent, mod])

    from . import mathmodule
    dst_type = mathmodule.binop_type(typesystem, base, exponent)
    result = mathmodule.infer_math_call(typesystem, node, base, exponent, mod)
    if dst_type.is_int:
        # TODO: Implement pow(int) in llvmmath
        return nodes.CoercionNode(result, dst_type)
    return result

@register_builtin((1, 2))
def round_(typesystem, node, number, ndigits):
    argtype = get_type(number)

    if len(node.args) == 1 and argtype.is_int:
        # round(myint) -> float(myint)
        return nodes.CoercionNode(node.args[0], double)

    if argtype.is_float or argtype.is_int:
        dst_type = double
    else:
        dst_type = object_
        node.args[0] = nodes.CoercionNode(node.args[0], object_)

    node.variable = Variable(dst_type)
    return node # nodes.CoercionNode(node, double)

def minmax(typesystem, args, op):
    if len(args) < 2:
        return

    res = args[0]
    for arg in args[1:]:
        lhs_type = get_type(res)
        rhs_type = get_type(arg)
        res_type = typesystem.promote(lhs_type, rhs_type)
        if lhs_type != res_type:
            res = nodes.CoercionNode(res, res_type)
        if rhs_type != res_type:
            arg = nodes.CoercionNode(arg, res_type)

        lhs_temp = nodes.TempNode(res_type)
        rhs_temp = nodes.TempNode(res_type)
        res_temp = nodes.TempNode(res_type)
        lhs = lhs_temp.load(invariant=True)
        rhs = rhs_temp.load(invariant=True)
        expr = ast.IfExp(ast.Compare(lhs, [op], [rhs]), lhs, rhs)
        body = [
            ast.Assign([lhs_temp.store()], res),
            ast.Assign([rhs_temp.store()], arg),
            ast.Assign([res_temp.store()], expr),
        ]
        res = nodes.ExpressionNode(body, res_temp.load(invariant=True))

    return res

@register_builtin(None)
def min_(typesystem, node, *args):
    return minmax(typesystem, args, ast.Lt())

@register_builtin(None)
def max_(typesystem, node, *args):
    return minmax(typesystem, args, ast.Gt())

@register_builtin(0)
def globals_(typesystem, node):
    return typesystem.dict_of_obj
    # return nodes.ObjectInjectNode(func.__globals__)

@register_builtin(0)
def locals_(typesystem, node):
    raise error.NumbaError("locals() is not supported in numba functions")

@register_builtin(1)
def ord_(typesystem, node, expr):
    type = get_type(expr)
    if type.is_int and type.typename in ("char", "uchar"):
        return nodes.CoercionNode(expr, int_)
    elif type.is_string:
        # TODO:
        pass

@register_builtin(1)
def chr_(typesystem, node, expr):
    type = get_type(expr)
    if type.is_int:
        return nodes.CoercionNode(expr, char)



# -*- coding: utf-8 -*-

"""
Default rules for the typing of constants.
"""

from __future__ import print_function, division, absolute_import

import math
import types
import ctypes
from functools import partial

import numba.typesystem
from numba.typesystem import itypesystem, numpy_support
from numba import numbawrapper

from numba.support.ctypes_support import is_ctypes, from_ctypes_value
from numba.support import cffi_support

import numpy as np

#------------------------------------------------------------------------
# Class -> Type
#------------------------------------------------------------------------

def get_typing_defaults(u):
    """
    Get a simple table mapping Python classes to types.

    :param u: The type universe
    """
    typing_defaults = {
        float: u.double,
        bool: u.bool_,
        complex: u.complex128,
        str: u.string_,
    }
    return typing_defaults

#------------------------------------------------------------------------
# Class -> pyval -> Type
#------------------------------------------------------------------------

def get_default_typing_rules(u, typeof, promote):
    """
    Get a table mapping Python classes to handlers (value -> type)

    :param u: The type universe
    """

    table = {}
    def register(*classes):
        def dec(func):
            for cls in classes:
                table[cls] = lambda u, value: func(value)
            return func
        return dec

    @register(int, long)
    def type_int(value):
        if abs(value) < 1:
            bits = 0
        else:
            bits = math.ceil(math.log(abs(value), 2))

        if bits < 32:
            return u.int_
        elif bits < 64:
            return u.int64
        else:
            raise ValueError("Cannot represent %s as int32 or int64", value)

    @register(np.ndarray)
    def type_ndarray(value):
        if isinstance(value, np.ndarray):
            dtype = numpy_support.map_dtype(value.dtype)
            return u.array(dtype, value.ndim)
                           #is_c_contig=value.flags['C_CONTIGUOUS'],
                           #is_f_contig=value.flags['F_CONTIGUOUS'])

    @register(tuple, list, dict)
    def type_container(value):
        assert isinstance(value, (tuple, list, dict))

        if isinstance(value, dict):
            key_type = type_container(value.keys(), promote, typeof)
            value_type = type_container(value.values(), promote, typeof)
            return u.dict_(key_type, value_type, size=len(value))

        if isinstance(value, tuple):
            container_type = u.tuple_
        else:
            container_type = u.list_

        if 0 < len(value) < 30:
            # Figure out base type if the container is not too large
            # base_type = reduce(promote, (typeof(child) for child in value))
            ty = typeof(value[0])
            if all(typeof(child) == ty for child in value):
                base_type = ty
            else:
                base_type = u.object_
        else:
            base_type = u.object_

        return container_type(base_type, size=len(value))

    register(np.dtype)(lambda value: u.numpy_dtype(numpy_support.map_dtype(value)))
    register(types.ModuleType)(lambda value: u.module(value))
    register(itypesystem.Type)(lambda value: u.meta(value))

    return table

def get_constant_typer(universe, typeof, promote):
    """
    Get a function mapping values to types, which returns None if unsuccessful.
    """
    typetable = get_typing_defaults(universe)
    handler_table = get_default_typing_rules(universe, typeof, promote)
    return itypesystem.ConstantTyper(universe, typetable, handler_table).typeof

#------------------------------------------------------------------------
# Constant matching ({ pyval -> bool : pyval -> Type })
#------------------------------------------------------------------------

# TODO: Make this a well-defined (easily overridable) matching table
# E.g. { "numpy" : { is_numpy : get_type } }

def is_dtype_constructor(value):
    return isinstance(value, type) and issubclass(value, np.generic)

def is_numpy_scalar(value): 
    return isinstance(value, np.generic)

def is_registered(value):
    from numba.type_inference import module_type_inference
    return module_type_inference.is_registered(value)

def from_ctypes(value, u):
    result = from_ctypes_value(value)
    if result.is_function:
        pointer = ctypes.cast(value, ctypes.c_void_p).value
        return u.pointer_to_function(value, pointer, result)
    else:
        return result

def from_cffi(value, u):
    signature = cffi_support.get_signature(value)
    pointer = cffi_support.get_pointer(value)
    return u.pointer_to_function(value, pointer, signature)

def from_typefunc(value, u):
    from numba.type_inference import module_type_inference
    result = module_type_inference.module_attribute_type(value)
    if result is not None:
        return result
    else:
        return u.known_value(value)

is_numba_exttype = lambda value: hasattr(type(value), '__numba_ext_type')
is_NULL = lambda value: value is numba.NULL
is_autojit_func = lambda value: isinstance(
    value, numbawrapper.NumbaSpecializingWrapper)

def get_default_match_table(u):
    """
    Get a matcher table: { (type -> bool) : (value -> type) }
    """
    table = {
        is_NULL:
            lambda value: numba.typesystem.null,
        is_dtype_constructor:
            lambda value: numba.typesystem.from_numpy_dtype(np.dtype(value)),
        is_numpy_scalar:
            lambda value: numpy_support.map_dtype(value.dtype),
        is_ctypes:
            lambda value: from_ctypes(value, u),
        cffi_support.is_cffi_func:
            lambda value: from_cffi(value, u),
        is_numba_exttype:
            lambda value: getattr(type(value), '__numba_ext_type'),
        numbawrapper.is_numba_wrapper:
            lambda value: u.jit_function(value),
        is_autojit_func:
            lambda value: u.autojit_function(value),
        is_registered:
            lambda value: from_typefunc(value, u),
    }

    return table

def find_match(matchtable, value):
    for matcher, typefunc in matchtable.iteritems():
        if matcher(value):
            result = typefunc(value)
            assert result is not None
            return result

    return None

#------------------------------------------------------------------------
# Typeof
#------------------------------------------------------------------------

def object_typer(universe, value):
    return universe.object_

def find_first(callables, value):
    for callable in callables:
        result = callable(value)
        if result is not None:
            return result

    assert False, (callables, value)

def get_default_typeof(universe, promote):
    typeof1 = get_constant_typer(universe, lambda value: typeof(value), promote)
    typeof2 = partial(find_match, get_default_match_table(universe))
    typeof3 = partial(object_typer, universe)
    typeof = partial(find_first, [typeof1, typeof2, typeof3])
    return typeof



from numba import *
import numpy as np

@jit(object_(object_[:, :]))
def func(A):
    for x in xrange(2):
        for y in xrange(2):
            items = A[x,y]
            if items == None:
                continue
            for item in items:
                print(item)

    return A


# -*- coding: utf-8 -*-

"""
Kinds for numba types.
"""

from __future__ import print_function, division, absolute_import

#------------------------------------------------------------------------
# Type Kinds
#------------------------------------------------------------------------

# Low level kinds

KIND_VOID       = "void"
KIND_INT        = "int"
KIND_FLOAT      = "float"
KIND_COMPLEX    = "complex"
KIND_FUNCTION   = "function"
KIND_ARRAY      = "array"
KIND_POINTER    = "pointer"
KIND_NULL       = "null"
KIND_CARRAY     = "carray"
KIND_STRUCT     = "struct"

# High-level Numba kinds

KIND_BOOL       = "bool"
KIND_OBJECT     = "object"
KIND_EXTTYPE    = "exttype"
KIND_NONE       = "none"



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba import autojit, jit

@autojit
def closure_modulo(a, b):
    @jit('int32()')
    def foo():
        return a % b
    return foo()


def test_closure_modulo():
    assert closure_modulo(100, 48) == 4

if __name__ == '__main__':
    test_closure_modulo()



# -*- coding: utf-8 -*-

"""
Visualize a CFG.
"""

from __future__ import print_function, division, absolute_import

import os
import ast
import textwrap

from numba import void
from numba import utils
from numba.viz import graphviz

# ______________________________________________________________________
# Utilities

def cf_from_source(source, func_globals):
    "Render the SSA graph given python source code"
    from numba import pipeline
    from numba import environment

    mod = ast.parse(source)
    func_ast = mod.body[0]

    env = environment.NumbaEnvironment.get_environment()
    func_env, _ = pipeline.run_pipeline2(env, None, func_ast, void(),
                                         pipeline_name='cf',
                                         function_globals=dict(func_globals))
    return func_env.symtab, func_env.flow #func_env.cfg

# ______________________________________________________________________
# Adaptor

class CFGGraphAdaptor(graphviz.GraphAdaptor):
    def children(self, node):
        return node.children

# ______________________________________________________________________
# Renderer

def fmtnode(node):
    if isinstance(node, ast.Assign):
        return "%s = %s" % (node.targets, node.value)
    else:
        return str(node)

fmtnode = utils.pformat_ast # ast.dump # str

class CFGGraphRenderer(graphviz.GraphRenderer):

    def render(self, node):
        return "%s\n%s" % (node, "; ".join(map(fmtnode, node.body)))

# ______________________________________________________________________
# Entry Points

def render_cfg(cfflow, output_file):
    "Render the SSA graph given the flow.CFGFlow and the symbol table"
    graphviz.render(cfflow.blocks, output_file,
                    CFGGraphAdaptor(), CFGGraphRenderer())

def render_cfg_from_source(source, output_file, func_globals=()):
    "Render the SSA graph given python source code"
    symtab, cfflow = cf_from_source(source, func_globals)
    render_cfg(cfflow, output_file)

# ______________________________________________________________________
# Test

if __name__ == '__main__':
    source = textwrap.dedent("""
        def func():
            # x_0
            x = 0 # x_1
            # x_2
            for i in range(10):
                if i < 5:
                    x = i # x_3
                # x_4
                x = x + i # x_5

            y = x
            x = i # x_6
    """)
    render_cfg_from_source(source, os.path.expanduser("~/cfg.dot"))


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from .llvm_testutils import *

def build_expr(type, op):
    out, v1, v2 = vars = build_vars(type, type, type)
    expr = b.assign(out, b.binop(type, op, v1, v2))
    return vars, expr

def build_kernel(specialization_name, ndim, type, op, **kw):
    vars, expr = build_expr(minitypes.ArrayType(type, ndim, **kw), op)
    func = MiniFunction(specialization_name, vars, expr, '%s_%s_%s' % (specialization_name, type.name, op))
    return func

comparison_operators = ['<', '<=', '>', '>=', '==', '!=']
arithmetic_operators = ['+', '-', '*', '/', '%'] # + ['**'] + # + comparison_operators
#bitwise_operators = ['<<', '>>', '|', '^', '&']
bitwise_operators = ['|', '^', '&']

a = np.random.random_sample((10, 20))

def _impl(type, op, x, y):
    func = build_kernel('strided', 2, type, op)

    dtype = minitypes.map_minitype_to_dtype(type)
    x = x.astype(dtype)
    y = y.astype(dtype)

    numpy_result = eval('a %s b' % (op,), {'a': x, 'b': y})
    our_result = func(x, y)
    assert np.all(numpy_result == our_result)

@parametrize(type=[short, int32, int64, float_, double], op=arithmetic_operators)
def test_arithmetic_operators(type, op):
    x = a
    y = np.arange(1, 10 * 20 + 1).reshape(10, 20)
    _impl(type, op, x, y)

@parametrize(type=[short, int32, int64], op=bitwise_operators)
def test_bitwise_operators(type, op):
    _impl(type, op, a * 100, a * 10)




# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba import error
from numba.typesystem import numpy_support
from numba.typesystem import integral, floating, complextypes, unsigned_integral
from numba.typesystem.kinds import *

import numpy as np

#------------------------------------------------------------------------
# Promotion on kind
#------------------------------------------------------------------------

table = {
    (KIND_POINTER, KIND_INT):   KIND_POINTER,
    (KIND_INT, KIND_POINTER):   KIND_POINTER,
    (KIND_POINTER, KIND_NULL):  KIND_POINTER,
    (KIND_NULL, KIND_POINTER):  KIND_POINTER,
    (KIND_OBJECT, KIND_OBJECT): KIND_OBJECT,
    (KIND_BOOL, KIND_BOOL):     KIND_BOOL,
}

def promote_from_table(table, u, promote, type1, type2):
    result = table.get((type1.kind, type2.kind))
    if result is not None:
        return { type1.kind: type1, type2.kind: type2}[result]
    return None

#------------------------------------------------------------------------
# Numeric promotion
#------------------------------------------------------------------------

def find_type_of_size(size, typelist):
    for type in typelist:
        if type.itemsize == size:
            return type

    assert False, "Type of size %d not found: %s" % (size, typelist)

def promote_int(u, promote, type1, type2):
    """
    Promote two integer types. We have to remain portable at this point, e.g.
    promote_int(Py_ssize_t, int) should be Py_ssize_t, and not int64 or int32.
    """
    t1, t2 = type1.get_dtype(), type2.get_dtype()
    return numpy_support.map_dtype(np.result_type(t1, t2))

    # size = max(type1.itemsize, type2.itemsize)
    # if type1.signed and type2.signed:
    #     return find_type_of_size(size, integral)
    #
    # if type1.signed:
    #     unsigned = type1
    #     other = type2
    # else:
    #     unsigned = type2
    #     other = type1
    #
    # if (type1.itemsize == size == type2.itemsize or not other.signed or
    #         other.itemsize <= size):
    #     return find_type_of_size(size, unsigned_integral)
    # else:
    #     return find_type_of_size(size, integral)

def promote_numeric(u, promote, type1, type2):
    "Promote two numeric types"
    ranks = ["bool", "int", "float", "complex"]
    type = max([type1, type2], key=lambda type: ranks.index(type.kind))
    size = max(type1.itemsize, type2.itemsize)
    if type.is_complex:
        return find_type_of_size(size, complextypes)
    elif type.is_float:
        return find_type_of_size(size, floating)
    else:
        assert type.is_int
        return promote_int(u, promote, type1, type2)

#------------------------------------------------------------------------
# Array promotion
#------------------------------------------------------------------------

def promote_arrays(u, promote, type1, type2):
    "Promote two array types in an expression to a new array type"
    equal_ndim = type1.ndim == type2.ndim
    return u.array(promote(type1.dtype, type2.dtype),
                   ndim=max((type1.ndim, type2.ndim)),
                   is_c_contig=(equal_ndim and type1.is_c_contig and
                                type2.is_c_contig),
                   is_f_contig=(equal_ndim and type1.is_f_contig and
                                type2.is_f_contig))

def promote_array_and_other(u, promote, type1, type2):
    if type1.is_array:
        array_type = type1
        other_type = type2
    else:
        array_type = type2
        other_type = type1

    if other_type.is_object and not array_type.dtype.is_object:
        # Make sure that (double[:], object_) -> object_
        return u.object_

    dtype = promote(array_type.dtype, other_type)
    return u.array(dtype, array_type.ndim)

#------------------------------------------------------------------------
# Default type promotion
#------------------------------------------------------------------------

class DefaultPromoter(object):

    def __init__(self, universe, promotion_table):
        self.universe = universe
        self.promotion_table = promotion_table

    def promote(self, type1, type2):
        "Promote two arbitrary types"
        if type1 == type2:
            return type1

        u = self.universe
        args = u, self.promote, type1, type2
        result = promote_from_table(self.promotion_table, *args)

        if result is not None:
            return result
        elif type1.is_numeric and type2.is_numeric:
            return promote_numeric(*args)
        elif type1.is_array and type2.is_array:
            return promote_arrays(*args)
        elif type1.is_array or type2.is_array:
            return promote_array_and_other(*args)
        elif (type1, type2) in [(u.string_, u.char.pointer()),
                                (u.char.pointer(), u.string_)]:
            return u.string_
        elif type1.is_object or type2.is_object:
            return u.object_
        else:
            raise error.UnpromotableTypeError((type1, type2))

def have_properties(type1, type2, property1, property2):
    """
    Return whether the two types satisfy the two properties:

    >>> have_properties(int32, int32.pointer(), "is_pointer", "is_int")
    True
    """
    type1_p1 = getattr(type1, property1)
    type1_p2 = getattr(type1, property2)
    type2_p1 = getattr(type2, property1)
    type2_p2 = getattr(type2, property2)

    if (type1_p1 and type2_p2) or (type1_p2 and type2_p1):
        if type1_p1:
            return type1
        else:
            return type2
    else:
        return None

#------------------------------------------------------------------------
# Promote
#------------------------------------------------------------------------

def get_default_promoter(universe):
    return DefaultPromoter(universe, table).promote